import numpy as np
import pytest

from pydrobert.speech import scales


@pytest.fixture(
    params=[
        scales.LinearScaling(10),
        scales.OctaveScaling(19),
        scales.MelScaling(),
        scales.BarkScaling(),
    ],
    ids=["linear", "octave", "mel", "bark"],
    scope="module",
)
def scaling_function(request):
    return request.param


def test_scales_invertible(scaling_function):
    for hertz in range(20, 8000):
        scale = scaling_function.hertz_to_scale(hertz)
        assert np.isclose(
            hertz, scaling_function.scale_to_hertz(scale)
        ), "Inverse not equal to orig for {} at {}".format(scaling_function, hertz,)



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Submodule for corpus iterators"""


from itertools import cycle
from typing import TypeVar, Type

from pydrobert.speech.post import PostProcessor
from pydrobert.speech.alias import alias_factory_subclass_from_arg


__all__ = ["post_process_wrapper"]

T = TypeVar("T", covariant=True)


def post_process_wrapper(cls: Type[T]) -> Type[T]:
    """Wrap a pydrobert-kaldi Data object for post-processing

    This function returns a class that wraps the `cls` class, performing some
    post-processing after batching
    """

    class _Wrapper(cls):
        def __init__(self, table, *additional_tables, **kwargs):
            postprocessors = kwargs.pop("postprocessors", dict())
            if not hasattr(postprocessors, "get"):
                postprocessors = {0: postprocessors}
            for key, value in list(postprocessors.items()):
                value = tuple(
                    alias_factory_subclass_from_arg(PostProcessor, postprocessor)
                    for postprocessor in value
                )
                postprocessors[key] = value
            self.postprocessors = postprocessors
            postprocess_axis = kwargs.pop("postprocess_axis", -1)
            if not hasattr(postprocess_axis, "__len__"):
                postprocess_axis = (postprocess_axis,)
            if not hasattr(postprocess_axis, "get"):
                post_dict = dict()
                for key in postprocessors:
                    post_dict[key] = postprocess_axis
                postprocess_axis = post_dict
            self.postprocess_axis = postprocess_axis
            super(_Wrapper, self).__init__(table, *additional_tables, **kwargs)

        def batch_generator(self, repeat=False):
            subsamples = self.num_sub != 1
            for batch in super(_Wrapper, self).batch_generator(repeat=repeat):
                if subsamples:
                    cur_batch = []
                    for sub_batch_idx, sub_batch in enumerate(batch):
                        for postprocessor, axis in zip(
                            self.postprocessors.get(sub_batch_idx, tuple()),
                            cycle(self.postprocess_axis.get(sub_batch_idx, tuple())),
                        ):
                            sub_batch = postprocessor.apply(
                                sub_batch, axis=axis, in_place=True
                            )
                        cur_batch.append(sub_batch)
                    yield tuple(cur_batch)
                else:
                    for postprocessor, axis in zip(
                        self.postprocessors[0], cycle(self.postprocess_axis[0])
                    ):
                        batch = postprocessor.apply(batch, axis=axis, in_place=True)
                    yield batch

    _Wrapper.__doc__ = cls.__doc__ + post_process_wrapper.WRAPPED_DATA_DOC
    return _Wrapper


post_process_wrapper.WRAPPED_DATA_DOC = """
This class has been wrapped using
:func:`pydrobert.speech.post_process_wrapper`. A set of additional
parameters has been introduced:

Additional Parameters
---------------------
postprocessors : sequence or mapping, optional
    Some number of :class:`pydrobert.speech.pre.PostProcessor` objects to apply to
    individual samples after batching. If a sequence is provided and
    ``num_sub == 1``, the sequence of preprocessors will be applied to the
    batch in the order they are presented. If ``num_sub > 1``, preprocessors
    will be applied to the first sub-batch per batch. If a mapping is
    provided, its values are sequences of preprocessors to apply to the
    sub-batch also indexed by the key
postprocess_axis : int or sequence or mapping, optional
    What axis/axis to apply postprocessors to. If an int, every postprocessor will be
    applied on that axis. If a sequence, the sequences of postprocessors specified in
    `postprocessors` will be applied one-to-one on the sequence of axes. If there are
    fewer postprocessors than axes, the axes will repeat. A mapping can be used to
    specify different sequences for different sub-batches. Defaults to ``-1``

Additional Attributes
---------------------
postprocessors : mapping
    The passed postprocessors, mapped to individual sub-batches
postprocess_axis : mapping
    The axis along which post-processing occurs
"""



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Visualization functions

Raises
------
ImportError
    This submodule requires :mod:`matplotlib`
"""

from itertools import cycle
from typing import Optional, Sequence, Tuple, Union

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import numpy as np

from matplotlib import pyplot as plt
from matplotlib import ticker
from matplotlib.axes import Axes
from matplotlib.colors import Colormap
from matplotlib.figure import Figure
from pydrobert.speech.filters import LinearFilterBank
from pydrobert.speech.compute import FrameComputer, LinearFilterBankFrameComputer
from pydrobert.speech.post import PostProcessor


__all__ = [
    "compare_feature_frames",
    "plot_frequency_response",
]


def plot_frequency_response(
    banks: Union[Sequence[LinearFilterBank], LinearFilterBank],
    axes: Optional[Axes] = None,
    dft_size: Optional[int] = None,
    half: Optional[bool] = None,
    title: Optional[str] = None,
    x_scale: Literal["hz", "ang", "bins"] = "hz",
    y_scale: Literal["dB", "power", "real", "imag", "both"] = "dB",
    cmap: Colormap = None,
) -> Figure:
    """Plot frequency response of filters in a filter bank

    Parameters
    ----------
    bank
    axes
        An :class:`Axes` object to plot on. Default is to generate a new figure
    dft_size
        The size of the Discrete Fourier Transform to plot. Defaults to
        ``max(max(bank.supports), 2 * bank.sampling_rate // min(bank.supports_hz)``
    half
        Whether to plot the half or full spectrum. Defaults to
        ``bank.is_real``
    title
        What to call the graph. The default is not to show a title
    x_scale
        The frequency coordinate scale along the x axis. Hertz
        (:obj:`'hz'`) is cycles/sec, angular frequency (:obj:`'ang'`) is
        radians/sec, and :obj:`'bins'` is the sample index within the DFT
    y_scale
        How to express the frequency response along the y axis. Decibels
        (:obj:`'dB'`) is the log of a ratio of the maximum quantity in the
        bank. The range between 0 and -20 decibels is displayed. Power
        spectrum (:obj:`'power'`) is the squared magnitude of the frequency
        response. :obj:`'real'` is the real part of the response,
        :obj:`'imag'` is the imaginary part of the response, and :obj:`'both'`
        displays both :obj:`'real'` and :obj:`'imag'` as separate lines
    cmap
        A :class:`Colormap` to pull colours from. Defaults to matplotlib's default
        colormap

    Returns
    -------
    fig : matplotlib.figure.Figure
        The containing figure
    """
    try:
        len(banks)
    except AttributeError:  # 1 bank
        banks = [banks]
    if not all(x.num_filts for x in banks):
        raise ValueError("Filter banks must have at least one filter to be visualized")
    if not all(x.sampling_rate == banks[0].sampling_rate for x in banks):
        raise ValueError("Banks must all have the same sampling rate")
    rate = banks[0].sampling_rate
    if cmap is None:
        cmap = plt.get_cmap()
    if dft_size is None:
        dft_size = max(
            int(
                max(
                    max(right - left for left, right in bank.supports),
                    2 * rate / min(right - left for left, right in bank.supports_hz),
                )
            )
            for bank in banks
        )
    if half is None:
        half = all(bank.is_real for bank in banks)
    if axes is None:
        fig, axes = plt.subplots()
    else:
        fig = axes.get_figure()
    colours = cmap.colors
    r_colours = list(cmap.colors)
    r_colours.reverse()
    responses_colours = []
    for bank, first_colour, second_color in zip(
        banks, cycle(colours), cycle(r_colours)
    ):
        responses_colours.extend(
            [
                (
                    bank.get_frequency_response(filt_idx, dft_size, half=half),
                    first_colour,
                    second_color,
                )
                for filt_idx in range(bank.num_filts)
            ]
        )
    if half:
        x = np.arange(
            (dft_size + dft_size % 2) // 2 + 1 - dft_size % 2, dtype=np.float32
        )
    else:
        x = np.arange(dft_size, dtype=np.float32)
    if x_scale in ("hz", "Hz", "hertz", "Hertz"):
        x_title = "Frequency (Hz)"
        x *= rate
        x /= dft_size
    elif x_scale in ("ang", "angle", "angular"):
        x_title = "Angular Frequency"
        x *= 2 * np.pi
        x /= dft_size
        axes.xaxis.set_major_locator(ticker.MultipleLocator(np.pi))
        axes.xaxis.set_minor_locator(ticker.AutoMinorLocator(2))
        axes.xaxis.set_major_formatter(ticker.FuncFormatter(_pi_formatter))
        axes.xaxis.set_minor_formatter(ticker.FuncFormatter(_pi_formatter))
    elif x_scale == "bins":
        x_title = "DFT Bin"
    else:
        raise ValueError("Invalid x_scale: {}".format(x_scale))
    if y_scale in ("db", "dB", "decibels"):
        y_title = "Log Ratio (dB)"
        # maximum abs. Get ripped
        max_abs = max(max(np.abs(response)) for response, _, _ in responses_colours)
        max_abs = np.log10(max(np.finfo(float).eps, max_abs))
        for filt_idx in range(len(responses_colours)):
            response, first_colour, second_colour = responses_colours[filt_idx]
            response = np.abs(response)
            response[response <= np.finfo(float).eps] = np.nan
            response[...] = 20 * (np.log10(response) - max_abs)
            # looks better than discontinuities
            response[np.isnan(response)] = -1e10
            responses_colours[filt_idx] = response, first_colour, second_colour
        y_max = 0
        y_min = -10
    elif y_scale in ("pow", "power"):
        y_title = "Power"
        y_min = 0
        y_max = 0
        for filt_idx in range(len(responses_colours)):
            response, first_colour, second_colour = responses_colours[filt_idx]
            response = np.abs(response) ** 2
            y_max = max(y_max, max(response))
            responses_colours[filt_idx] = response, first_colour, second_colour
        y_max *= 1.04
    elif y_scale in ("real", "imag", "imaginary", "both"):
        if y_scale == "real":
            y_title = "Real-value response"
        elif y_scale == "both":
            y_title = "Complex response"
        else:
            y_title = "Imaginary-value response"
        y_min = np.inf
        y_max = -np.inf
        new_responses_colours = []
        for response, first_colour, second_colour in responses_colours:
            if y_scale == "real":
                response = np.real(response)
            elif y_scale == "both":
                response_b = np.imag(response)
                response = np.real(response)
                y_max = max(y_max, max(response_b))
                y_min = min(y_min, min(response_b))
                new_responses_colours.append((response_b, second_colour, first_colour))
            else:
                response = np.imag(response)
            y_max = max(y_max, max(response))
            y_min = min(y_min, min(response))
            new_responses_colours.append((response, first_colour, second_colour))
        assert np.isfinite(y_min) and np.isfinite(y_max)
        y_max *= 0.96 if y_max < 0 else 1.04
        y_min *= 0.96 if y_min > 0 else 1.04
        del responses_colours
        responses_colours = new_responses_colours
    else:
        raise ValueError("Invalid y_scale: {}".format(y_scale))
    axes.set_xlim((0, max(x)))
    axes.set_ylim((y_min, y_max))
    if title:
        axes.set_title(title)
    axes.set_ylabel(y_title)
    axes.set_xlabel(x_title)
    for response, colour, _ in responses_colours:
        axes.plot(x, response, color=colour)
    # if y_scale == 'both':
    #     real_handle = lines.Line2D([], [], color=real_colour, label='Real')
    #     imag_handle = lines.Line2D([], [], color=imag_colour, label='Imag')
    #     axes.legend(handles=[real_handle, imag_handle])
    return fig


def _pi_formatter(val, _):
    num_halfpi = int(np.round(2 * val / np.pi))
    if np.isclose(num_halfpi * np.pi / 2, val):
        if not num_halfpi:
            return "0"
        elif num_halfpi == 1:
            return "\\u03C0 / 2"
        elif num_halfpi == -1:
            return "-\\u03C0 / 2"
        elif num_halfpi == 2:
            return "\\u03C0"
        elif num_halfpi == -2:
            return "-\\u03C0"
        elif num_halfpi % 2:
            return "{}\\u03C0 / 2".format(num_halfpi)
        else:
            return "{}\\u03C0".format(num_halfpi // 2)
    else:
        return ""


def compare_feature_frames(
    computers: Union[FrameComputer, Sequence[FrameComputer]],
    signal: np.ndarray,
    axes: Optional[int] = None,
    figure_height: float = None,
    figure_width: float = None,
    plot_titles: Tuple[str, ...] = None,
    positions: Tuple[Union[int, Tuple[int, int]], ...] = None,
    post_ops: Optional[Union[PostProcessor, Sequence[PostProcessor]]] = None,
    title: Optional[str] = None,
    **kwargs
) -> Figure:
    """Compare features from frame computers via spectrogram-like heat map

    Direct comparison of :class:`FrameComputer` objects is possible because all
    subclasses of this abstract data type share a common interpretation of frame
    boundaries (according to :func:`FrameComputer.frame_style`).

    Additional keyword args will be passed to the plotting routine.

    Parameters
    ----------
    computers
        One or more frame computers to compare
    signal
        A 1D array of the raw speech. Assumed to be valid with respect to computer
        settings (e.g. sample rate).
    axes
        By default, this function creates a new figure and subplots. Setting one
        `axes` value for every `computers` value will plot feature representations from
        `computers` into each ordered :class:`Axes`. If `axes` do not belong to the same
        figure, a :class:`ValueError` will be raised
    figure_height
        If a new figure is created, this sets the figure height (in inches). This value
        is determined dynamically according to `figure_width` by default. A
        :class:`ValueError` will be raised if both `figure_height` and `axes` are set
    figure_width
        If a new figure is created, this set the figure width (in inches). This value
        defaults to 3.33 inches if all subplots are positioned vertically, and to 7
        inches if there are at least two columns of plots. A :class:`ValueError` will be
        raised if both `figure_width` and `axes` are set
    plot_titles
        An ordered list of strings specifying the titles of each subplot. The default is
        to not display subplot titles
    positions
        If a new figure is created, `positions` decides how the
        subplots should be positioned relative to one another. Can
        contain only ints (describing the position on only the row-axis)
        or pairs of ints (describing the row-col positions). Positions
        must be contiguous and start from index 0 or 0,0 (top or
        top-left). `positions` cannot be specified if `axes` is
        specified
    post_ops
        One or more post-processors to apply (in order) to each computed
        feature representation. If a simple list of post-processors is
        provided, each operation is applied to the default axis (the
        feature coefficient axis). To explicitly set the axis, pairs of
        ``(op, axis)`` can be specified in the list. No op is allowed
        to change the shape of the feature representation
        (e.g. :class:`Deltas`), or a :class:`ValueError` will be thrown
    title
        The title of the whole figure. Default is to display no title

    Returns
    -------
    fig : matplotlib.figure.Figure
        The containing figure
    """

    try:
        iter(computers)
    except TypeError:
        computers = (computers,)
    if not len(computers):
        raise ValueError("Expected at least one computer")
    if plot_titles is not None:
        try:
            iter(plot_titles)
        except TypeError:
            plot_titles = [plot_titles]
        if len(plot_titles) != len(computers):
            raise ValueError(
                "Expected {} plot titles, got {}".format(
                    len(computers), len(plot_titles)
                )
            )
    else:
        plot_titles = [None] * len(computers)
    if positions is not None:
        if len(computers) == 1 and positions not in (0, (0,), [0]):
            raise ValueError("Nonzero position specified for only one plot")
        elif axes is not None:
            raise ValueError("Cannot specify positions of predefined axes")
        elif len(positions) != len(computers):
            raise ValueError(
                "Expected {} positions, got {}".format(len(computers), len(positions))
            )
        if any(hasattr(p, "__iter__") for p in positions) and not all(
            len(p) == 1 for p in positions if hasattr(p, "__iter__")
        ):
            # expect 2-dimensional plot positioning
            if any(not hasattr(p, "__iter__") or len(p) != 2 for p in positions):
                raise ValueError("Expected all plot positions to be two-dimensional")
            row_set = set(p[0] for p in positions)
            col_set = set(p[1] for p in positions)
            row_len, col_len = max(row_set) + 1, max(col_set) + 1
            if row_set != set(r for r in range(row_len)) or col_set != set(
                c for c in range(col_len)
            ):
                raise ValueError("positions not contiguous")
            gs_args = (row_len, col_len)
        else:
            # expect 1-dimensional plot positioning. Using gridspec,
            # so have to add a column coordinate
            positions = tuple(
                (next(iter(p)), 0) if hasattr(p, "__iter__") else p for p in positions
            )
            row_set = set(p[0] for p in positions)
            row_len = max(row_set) + 1
            if row_set != set(r for r in range(row_len)):
                raise ValueError("positions not contiguous")
            gs_args = (row_len, 1)
    elif axes is None:
        # choose our own positions
        num_plots = len(computers)
        row_len = int(np.ceil(num_plots ** 0.5))
        col_len = row_len
        while col_len * row_len != num_plots:
            if col_len * row_len > num_plots and col_len > 1:
                row_len += 1
                col_len -= 1
            else:
                row_len -= 1
        gs_args = (row_len, col_len)
        positions = tuple(np.ndindex(gs_args))
    if figure_width is not None:
        if axes is not None:
            raise ValueError("Cannot specify figure width with predefined axes")
    elif axes is None:
        figure_width = 7.0 if gs_args[1] > 1 else 3.33
    if figure_height is not None:
        if axes is not None:
            raise ValueError("Cannot specify figure height with predefined axes")
    elif axes is None:
        figure_height = figure_width * 9 / 16 / gs_args[1] * gs_args[0]
    if post_ops is not None:
        try:
            iter(post_ops)
        except TypeError:
            post_ops = (post_ops,)
        if len(post_ops) == 2 and isinstance(post_ops[1], int):
            post_ops = (post_ops,)
    else:
        post_ops = []
    if axes is not None:
        try:
            iter(axes)
        except TypeError:
            axes = (axes,)
        if len(axes) != len(computers):
            raise ValueError(
                "Expected {} axes, got {}".format(len(computers), len(axes))
            )
        fig = axes[0].get_figure()
        for ax in axes[1:]:
            if ax.get_figure() != fig:
                raise ValueError("Axes do not share the same figure")
    else:
        fig = plt.figure(figsize=(figure_width, figure_height))
        if len(computers) == 1:
            axes = (fig.add_subplot(111),)
        else:
            axes = []
            sharey = all(
                isinstance(computer, LinearFilterBankFrameComputer)
                for computer in computers
            )
            gridspec = plt.GridSpec(gs_args[0], gs_args[1])
            for position in positions:
                if axes and sharey:
                    ax = fig.add_subplot(
                        gridspec[position], sharex=axes[0], sharey=axes[0]
                    )
                elif axes:
                    ax = fig.add_subplot(gridspec[position], sharex=axes[0])
                else:
                    ax = fig.add_subplot(gridspec[position])
                axes.append(ax)
    supremum_seconds = np.infty
    num_samples = len(signal)
    for idx, (computer, ax, plot_title) in enumerate(zip(computers, axes, plot_titles)):
        frame_length = computer.frame_length
        frame_shift = computer.frame_shift
        if computer.frame_style == "causal":
            pad_left = 0
        else:  # centered
            pad_left = (frame_length + 1) // 2 - 1
        total_len = num_samples + pad_left
        num_frames = max(0, (total_len - frame_length) // frame_shift + 1)
        # individual computers may choose to add a final frame by
        # padding. Since this behaviour is not guaranteed, we only
        # consider full frames
        if not num_frames:
            raise ValueError(
                "The computer indexed at {} is unable to generate "
                "a full frame from the signal".format(idx)
            )
        # we use frame shifts to specify bounds (frame length is likely
        # overlapping), with the exception of the last frame
        sample_bounds = np.arange(num_frames + 1, dtype=float) * frame_shift
        if pad_left:
            # r.h.s. bound half a frame shift to right of center (or
            # half frame right of center for last frame)
            # l.h.s. bound half the other way (or 0 for first frame)
            sample_bounds[1:-1] -= (frame_shift + 1) // 2 - 1
            sample_bounds[-1] = sample_bounds[-2] + pad_left
        else:
            # l.h.s bound leftmost idx of each frame
            # r.h.s. is l.h.s. plus frame shift (or frame length for
            # last frame)
            sample_bounds[-1] = sample_bounds[-2] + frame_length
        seconds_bounds = sample_bounds / computer.sampling_rate
        supremum_seconds = min(supremum_seconds, seconds_bounds[-1])
        feat_slice = [slice(None, num_frames), slice(None)]
        if isinstance(computer, LinearFilterBankFrameComputer):
            yscale_label = "Frequency (Hz)"
            bank = computer.bank
            num_coeffs = bank.num_filts
            if computer.includes_energy:
                feat_slice[-1] = slice(1, None)
            supports_hz = bank.supports_hz
            assert num_coeffs == len(supports_hz)
            centers_hz = tuple((left + right) / 2 for left, right in supports_hz)
            # supports may be overlapping or sparse. Instead of using
            # supports to directly specify boundaries, we use them as
            # weights to pick points between center frequencies (except
            # the first and last filters, which get to extend their
            # lower and higher bounds to their supports, respectively.
            feature_bounds = np.empty(num_coeffs + 1)
            feature_bounds[0] = max(0, supports_hz[0][0])
            feature_bounds[-1] = min(computer.sampling_rate / 2, supports_hz[-1][-1])
            for high_idx in range(1, num_coeffs):
                low_c = centers_hz[high_idx - 1]
                high_c = centers_hz[high_idx]
                low_s, high_s = supports_hz[high_idx - 1]
                assert high_c >= low_c
                split_c = low_c * (high_s / (low_s + high_s))
                split_c += high_c * (low_s / (low_s + high_s))
                feature_bounds[high_idx] = split_c
        else:
            # no idea how to handle. Just plot rectangular coefficients
            num_coeffs = computer.num_coeffs
            yscale_label = None
            feature_bounds = np.arange(num_coeffs + 1)
        features = computer.compute_full(signal)
        assert features.shape[0] >= num_frames
        assert features[feat_slice].shape[-1] == num_coeffs
        for post_op_idx, post_op in enumerate(post_ops):
            try:
                apply_axis = post_op[1]
                post_op = post_op[0]
            except TypeError:
                apply_axis = -1
            new_features = post_op.apply(features, axis=apply_axis)
            if new_features.shape != features.shape:
                raise ValueError(
                    "The post_op indexed at {} changed the shape of the"
                    "features".format(post_op_idx)
                )
            features = new_features
        ax.pcolormesh(seconds_bounds, feature_bounds, features[feat_slice].T, **kwargs)
        if plot_title is not None:
            ax.set_title(plot_title)
        ax.set_xlabel("Time (seconds)")
        if yscale_label:
            ax.set_ylabel(yscale_label)
    for ax in axes:
        ax.set_xlim((0, supremum_seconds))
    if title:
        fig.suptitle(title)
    return fig



import numpy as np
import pytest

from pydrobert.speech.corpus import post_process_wrapper
from pydrobert.speech.post import PostProcessor


class Sum(PostProcessor):
    """Sum over the given axis

    Container for numpy.sum. Big whoop.
    """

    aliases = {"sum"}

    def apply(self, features, axis=-1, in_place=False):
        return np.sum(features, axis=axis)


class Axis(PostProcessor):
    """Return the passed axis"""

    aliases = {"axis"}

    def apply(self, features, axis=-1, in_place=False):
        return axis


def test_post_processor_wrapper(temp_file_1_name, temp_file_2_name):
    kaldi_io = pytest.importorskip("pydrobert.kaldi.io")
    kaldi_corpus = pytest.importorskip("pydrobert.kaldi.io.corpus")
    Wrapped = post_process_wrapper(kaldi_corpus.SequentialData)
    feats = []
    labels = []
    alphabet = tuple("abcdefgh")  # the expurgated version
    with kaldi_io.open(
        "ark:" + temp_file_1_name, "dm", mode="w"
    ) as writer_1, kaldi_io.open("ark:" + temp_file_2_name, "tv", mode="w") as writer_2:
        for sample_idx in range(100):
            feat_sample = np.random.random((np.random.randint(1, 100), 20))
            label_sample = tuple(
                np.random.choice(alphabet, size=np.random.randint(1, 10)).tolist()
            )
            feats.append(feat_sample)
            labels.append(label_sample)
            writer_1.write("{:03d}".format(sample_idx), feat_sample)
            writer_2.write("{:03d}".format(sample_idx), label_sample)
    wrapped_1 = Wrapped(
        ("ark,s:" + temp_file_1_name, "dm"),
        ("ark,s:" + temp_file_2_name, "tv"),
        postprocessors=["sum"],
    )
    n_samples = 0
    for ex_feat, ex_lab, (act_feat, act_lab) in zip(feats, labels, wrapped_1):
        assert ex_lab == act_lab
        assert act_feat.shape == ex_feat.shape[:-1]
        assert np.allclose(np.sum(ex_feat, axis=-1), act_feat)
        n_samples += 1
    assert n_samples == len(feats)
    wrapped_2 = Wrapped(
        ("ark,s:" + temp_file_1_name, "dm"),
        ("ark,s:" + temp_file_2_name, "tv"),
        postprocessors={0: ["sum"], 1: ["axis"]},
        postprocess_axis=[0, -1],
    )
    n_samples = 0
    for ex_feat, (act_feat, act_lab) in zip(feats, wrapped_2):
        assert act_lab == 0
        assert ex_feat.shape[1:] == act_feat.shape
        assert np.allclose(np.sum(ex_feat, axis=0), act_feat)
        n_samples += 1
    assert n_samples == len(feats)
    wrapped_3 = Wrapped(
        ("ark,s:" + temp_file_1_name, "dm"),
        ("ark,s:" + temp_file_2_name, "tv"),
        postprocessors={0: ["sum", "sum"], 1: ["axis", "axis"]},
        postprocess_axis={0: [1], 1: [2, -1]},
        batch_pad_mode="constant",
        batch_size=5,
    )
    n_samples = 0
    for act_feat, act_lab in wrapped_3:
        assert act_lab == -1
        ex_feats = feats[n_samples : n_samples + 5]
        # sum axis 1 then sum axis 1 means we've retained the batch axis
        assert np.allclose(np.stack([np.sum(s) for s in ex_feats]), act_feat)
        n_samples += len(ex_feats)
    assert n_samples == len(feats)



import numpy as np
import pytest

from pydrobert.speech import post


@pytest.fixture(
    params=[np.float64, np.float32, np.int32, np.int16],
    ids=["f64", "f32", "i32", "i16"],
    scope="module",
)
def dtype(request):
    return request.param


@pytest.mark.parametrize("norm_var", [True, False])
@pytest.mark.parametrize(
    "buff",
    [
        x * np.random.randint(1, 100) + np.random.randint(-10, 10)
        for x in [
            np.random.random((100, 1)),
            np.random.random((1, 10)),
            np.random.random((5, 5)),
            np.random.random((10, 4, 3)),
        ]
    ],
)
def test_standardize_local(norm_var, buff, dtype):
    if np.allclose(buff, buff[:1].ravel()[0]):
        pytest.skip()
    buff = buff.astype(dtype)
    stand = post.Standardize(norm_var=norm_var)
    for axis in range(len(buff.shape)):
        buff_2 = buff.copy()
        other_axes = tuple(idx for idx in range(len(buff.shape)) if idx != axis)
        if sum(buff_2.shape[idx] for idx in other_axes) == len(other_axes):
            continue
        # we make sure that one sample along the target axis is
        # different from at least one other
        s_1 = [0] * len(buff_2.shape)
        s_2 = [-1] * len(buff_2.shape)
        s_1[axis] = slice(None)
        s_2[axis] = slice(None)
        s_1 = tuple(s_1)
        s_2 = tuple(s_2)
        buff_2[s_1] = buff_2[s_2] - 1
        assert not np.any(np.isclose(buff_2.std(axis=other_axes), 0))
        s_buff = stand.apply(buff_2, axis=axis)
        assert np.allclose(s_buff.mean(axis=other_axes), 0)
        assert not np.allclose(s_buff, 0)
        if norm_var:
            assert np.allclose(s_buff.var(axis=other_axes), 1), axis


@pytest.mark.parametrize("norm_var", [True, False])
@pytest.mark.parametrize(
    "buff",
    [
        np.random.random((5, 100)) * np.random.randint(1, 100, 100)
        + (np.random.randint(-10, 10, 100)),
        np.random.random((8, 1, 10)) * np.random.randint(1, 100, 10)
        + (+np.random.randint(-10, 10, 10)),
        np.random.random((3, 10, 20)) * np.random.randint(1, 100, 20)
        + (+np.random.randint(-10, 10, 20)),
        np.random.random((2, 50, 2, 3)) * np.random.randint(1, 100, 3)
        + (+np.random.randint(-10, 10, 3)),
    ],
)
def test_standardize_global(norm_var, buff, dtype):
    buff = buff.astype(dtype)
    if np.allclose(buff, buff[:1].ravel()[0]):
        pytest.skip()
    other_axes = tuple(range(len(buff.shape) - 1))
    if norm_var:
        # quick fix when ints misbehave and give zero variance
        if np.any(np.isclose(buff.std(axis=other_axes), 0)):
            buff = np.zeros(buff.shape, dtype=buff.dtype)
            buff[0, ...] = 1
    stand = post.Standardize(norm_var=norm_var)
    for feats in buff:
        stand.accumulate(feats)
    s_buff_1 = stand.apply(buff)
    assert np.allclose(s_buff_1.mean(axis=other_axes), 0)
    if norm_var:
        assert np.allclose(s_buff_1.var(axis=other_axes), 1)
    # ensure that we're using stored statistics. Otherwise, s_buff_2
    # will have mean zero but not match s_buff_1
    s_buff_2 = stand.apply(buff[0])
    assert np.allclose(s_buff_1[0], s_buff_2)


def test_standardize_write_read(temp_file_1_name):
    stand_1 = post.Standardize()
    x_1 = np.random.random((2, 3, 4))
    x_2 = np.random.random((1, 3, 5)) + np.random.randint(-10, 10)
    x_3 = np.random.random((3, 3, 3)) * 100 - np.random.randint(-10, 10)
    stand_1.accumulate(x_1, axis=1)
    stand_1.accumulate(x_2, axis=1)
    x_1_p_1 = stand_1.apply(x_1, axis=1)
    stand_1.save(temp_file_1_name)
    stand_1.accumulate(x_3, axis=1)
    x_1_p_2 = stand_1.apply(x_1, axis=1)
    assert not np.allclose(x_1_p_1, x_1_p_2)
    stand_1.save(temp_file_1_name)
    stand_2 = post.Standardize(temp_file_1_name, force_as="file")
    x_1_p_3 = stand_2.apply(x_1, axis=1)
    assert np.allclose(x_1_p_2, x_1_p_3)


@pytest.mark.parametrize(
    "buff",
    [
        np.random.random(10),
        np.random.random((2, 5)),
        np.random.random((3, 6, 4)),
        np.random.random((5, 4, 0, 0, 1)),
    ],
)
@pytest.mark.parametrize("concatenate", [True, False])
@pytest.mark.parametrize("num_deltas", list(range(5)))
def test_delta_shapes(buff, concatenate, num_deltas):
    for target_axis in range(len(buff.shape) + 1 - int(concatenate)):
        deltas = post.Deltas(
            num_deltas, concatenate=concatenate, target_axis=target_axis
        )
        for axis in range(len(buff.shape)):
            new_shape = list(buff.shape)
            if concatenate:
                new_shape[target_axis] *= num_deltas + 1
            else:
                new_shape.insert(target_axis, num_deltas + 1)
            assert deltas.apply(buff, axis=axis).shape == tuple(new_shape), buff.shape


class KaldiDeltas(object):
    """Replicate Kaldi delta logic for comparative purposes"""

    def __init__(self, num_deltas, window=2):
        self._scales = [np.ones(1, dtype=np.float64)]
        for last_idx in range(num_deltas):
            prev_scale = self._scales[last_idx]
            cur_scale = np.zeros(len(prev_scale) + window * 2, dtype=np.float64)
            prev_offset = (len(prev_scale) - 1) // 2
            cur_offset = prev_offset + window
            normalizer = 0
            for j in range(-window, window + 1):
                normalizer += j * j
                for k in range(-prev_offset, prev_offset + 1):
                    cur_scale[j + k + cur_offset] += j * prev_scale[k + prev_offset]
            cur_scale /= normalizer
            self._scales.append(cur_scale)

    def _process(self, r, features, out_row):
        features = features.astype(np.float64, copy=False)
        num_frames, feat_dim = features.shape
        assert len(out_row) == feat_dim * len(self._scales)
        for idx, scale in enumerate(self._scales):
            max_offset = (len(scale) - 1) // 2
            sub_row = out_row[idx * feat_dim : (idx + 1) * feat_dim]
            for j in range(-max_offset, max_offset + 1):
                offset_frame = r + j
                if offset_frame < 0:
                    offset_frame = 0
                elif offset_frame >= num_frames:
                    offset_frame = num_frames - 1
                sub_row += scale[j + max_offset] * features[offset_frame]

    def apply(self, features):
        assert len(features.shape) == 2
        out = np.zeros(
            (features.shape[0], features.shape[1] * len(self._scales)), dtype=np.float64
        )
        for r, out_row in enumerate(out):
            self._process(r, features, out_row)
        return out.astype(features.dtype, copy=False)


@pytest.mark.parametrize(
    "buff",
    [np.random.random((1, 3)), np.random.random((3, 1)), np.random.random((20, 50))],
)
@pytest.mark.parametrize("num_deltas", list(range(5)))
@pytest.mark.parametrize("window", list(range(1, 6)))
def test_compare_to_kaldi(buff, num_deltas, window, dtype):
    buff = buff.astype(dtype)
    deltas = post.Deltas(
        num_deltas, concatenate=True, context_window=window, target_axis=1
    )
    kaldi_deltas = KaldiDeltas(num_deltas, window)
    delta_res = deltas.apply(buff, axis=0)
    kaldi_res = kaldi_deltas.apply(buff)
    assert np.allclose(delta_res, kaldi_res)


@pytest.mark.parametrize("transpose", [True, False])
def test_stack_2dim(transpose):
    buff = np.arange(30).reshape(10, 3)
    exp = buff[:9].reshape(3, 9)
    axis, time_axis = 1, 0
    if transpose:
        axis, time_axis = time_axis, axis
        buff, exp = buff.T, exp.T
    stack = post.Stack(3, time_axis=time_axis)
    act = stack.apply(buff, axis)
    assert (exp == act).all()


@pytest.mark.parametrize("pad_mode", [None, "edge"])
def test_stack_3dim(pad_mode):
    buff = np.arange(20).reshape(5, 2, 2)
    stack = post.Stack(2, pad_mode=pad_mode)
    exp = np.array(
        [
            [[0, 1, 4, 5], [2, 3, 6, 7]],
            [[8, 9, 12, 13], [10, 11, 14, 15]],
            [[16, 17, 16, 17], [18, 19, 18, 19]],
        ]
    )
    act = stack.apply(buff)
    if pad_mode is None:
        assert act.shape == (2, 2, 4)
        assert np.all(act == exp[:2])
    else:
        assert act.shape == exp.shape
        assert np.all(act == exp)




import os
import warnings

from tempfile import NamedTemporaryFile, mkdtemp
from shutil import rmtree

import pytest

try:
    from pydrobert.kaldi import KaldiLocaleWarning  # type: ignore

    warnings.filterwarnings("ignore", category=KaldiLocaleWarning)
except ImportError:
    pass

warnings.simplefilter("error")
# # annoying scipy errors. Not mah fault!
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")
warnings.filterwarnings("ignore", category=ImportWarning)
warnings.filterwarnings("ignore", category=DeprecationWarning)


# fixtures
@pytest.fixture
def temp_file_1_name():
    temp = NamedTemporaryFile(suffix="_1", delete=False)
    temp.close()
    yield temp.name
    os.remove(temp.name)


@pytest.fixture
def temp_file_2_name():
    temp = NamedTemporaryFile(suffix="_2", delete=False)
    temp.close()
    yield temp.name
    os.remove(temp.name)


@pytest.fixture
def temp_dir():
    dir_name = mkdtemp()
    yield dir_name
    rmtree(dir_name)



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Filters and filter banks"""


import abc
from typing import Mapping, Optional, Tuple, Union

import numpy as np

import pydrobert.speech.config as config

from pydrobert.speech.alias import AliasedFactory, alias_factory_subclass_from_arg
from pydrobert.speech.scales import MelScaling
from pydrobert.speech.scales import ScalingFunction
from pydrobert.speech.util import angular_to_hertz
from pydrobert.speech.util import hertz_to_angular

__all__ = [
    "BartlettWindow",
    "BlackmanWindow",
    "ComplexGammatoneFilterBank",
    "GaborFilterBank",
    "GammaWindow",
    "HammingWindow",
    "HannWindow",
    "LinearFilterBank",
    "TriangularOverlappingFilterBank",
    "WindowFunction",
]

# banks


class LinearFilterBank(AliasedFactory):
    """A collection of linear, time invariant filters

    A :class:`LinearFilterBank` instance is expected to provide factory methods for
    instantiating a fixed number of LTI filters in either the time or frequency domain.
    Filters should be organized lowest frequency first.
    """

    @abc.abstractproperty
    def is_real(self) -> bool:
        """bool : Whether the filters are real or complex"""
        pass

    @abc.abstractproperty
    def is_analytic(self) -> bool:
        """bool : Whether the filters are (approximately) analytic
        
        An analytic signal has no negative frequency components. A real signal cannot
        be analytic.
        """
        pass

    @abc.abstractproperty
    def is_zero_phase(self) -> bool:
        """bool : Whether the filters are zero phase or not

        Zero phase filters are even functions with no imaginary part in the fourier
        domain. Their impulse responses center around 0.
        """
        pass

    @abc.abstractproperty
    def num_filts(self) -> int:
        """int : Number of filters in the bank"""
        pass

    @abc.abstractproperty
    def sampling_rate(self) -> float:
        """float : Number of samples in a second of a target recording"""
        pass

    @abc.abstractproperty
    def supports_hz(self) -> Tuple[Tuple[float, float], ...]:
        """tuple : Boundaries of effective support of filter freq responses, in Hz.

        Returns a tuple of length `num_filts` containing pairs of floats of the low and
        high frequencies. Frequencies outside the span have a response of approximately
        (with magnitude up to :obj:`pydrobert.speech.EFFECTIVE_SUPPORT_SIGNAL`) zero.

        The boundaries need not be tight, i.e. the region inside the boundaries could be
        zero. It is more important to guarantee that the region outside the boundaries
        is approximately zero.

        The boundaries ignore the Hermitian symmetry of the filter if it is real. Bounds
        of ``(10, 20)`` for a real filter imply that the region ``(-20, -10)`` could
        also be nonzero.

        The user is responsible for adjusting the for the periodicity induced by
        sampling. For example, if the boundaries are ``(-5, 10)`` and the filter is
        sampled at 15Hz, then all bins of an associated DFT could be nonzero.
        """
        pass

    @abc.abstractproperty
    def supports(self) -> Tuple[Tuple[float, float], ...]:
        """tuple : Boundaries of effective support of filter impulse resps, in samples

        Returns a tuple of length `num_filts` containing pairs of integers of the first
        and last (effectively) nonzero samples.

        The boundaries need not be tight, i.e. the region inside the boundaries could be
        zero. It is more important to guarantee that the region outside the boundaries
        is approximately zero.

        If a filter is instantiated using a buffer that is unable to fully contain the
        supported region, samples will wrap around the boundaries of the buffer.

        Noncausal filters will have start indices less than 0. These samples will wrap
        to the end of the filter buffer when the filter is instantiated.
        """
        pass

    @property
    def supports_ms(self) -> Tuple[Tuple[float, float], ...]:
        """tuple : Boundaries of effective support of filter impulse resps, in ms"""
        return tuple(
            (s[0] * 1000 / self.sampling_rate, s[1] * 1000 / self.sampling_rate,)
            for s in self.supports
        )

    @abc.abstractmethod
    def get_impulse_response(self, filt_idx: int, width: int) -> np.ndarray:
        """Construct filter impulse response in a fixed-width buffer

        Construct the filter in the time domain.

        Parameters
        ----------
        filt_idx
            The index of the filter to generate. Less than `num_filts`
        width
            The length of the buffer, in samples. If less than the support of the
            filter, the filter will alias.

        Returns
        -------
        ir : np.ndarray
            1D float64 or complex128 numpy array of length `width`
        """
        pass

    @abc.abstractmethod
    def get_frequency_response(
        self, filt_idx: int, width: int, half: bool = False
    ) -> np.ndarray:
        """Construct filter frequency response in a fixed-width buffer

        Construct the 2pi-periodized filter in the frequency domain. Zero-phase filters
        are returned as 8-byte float arrays. Otherwise, they will be 16-byte complex
        floats.

        Parameters
        ----------
        filt_idx
            The index of the filter to generate. Less than `num_filts`
        width
            The length of the DFT to output
        half
            Whether to return only the DFT bins between ``[0,pi]``

        Returns
        -------
        fr : np.ndarray
            If `half` is :obj:`False`, returns a 1D float64 or complex128 numpy array of
            length `width`. If `half` is :obj:`True` and `width` is even, the returned
            array is of length ``width // 2 + 1``. If `width` is odd, the returned array
            is of length ``(width + 1) // 2``.
        """
        pass

    @abc.abstractmethod
    def get_truncated_response(
        self, filt_idx: int, width: int
    ) -> Tuple[int, np.ndarray]:
        """Get nonzero region of filter frequency response

        Many filters will be compactly supported in frequency (or approximately so).
        This method generates a tuple `(bin_idx, buf)` of the nonzero region.

        In the case of a complex filter, ``bin_idx + len(buf)`` may be greater than
        `width`; the filter wraps around in this case. The full frequency response can
        be calculated from the truncated response by:

        >>> bin_idx, trnc = bank.get_truncated_response(filt_idx, width)
        >>> full = numpy.zeros(width, dtype=trnc.dtype)
        >>> wrap = min(bin_idx + len(trnc), width) - bin_idx
        >>> full[bin_idx:bin_idx + wrap] = trnc[:wrap]
        >>> full[:len(trnc) - wrap] = tnc[wrap:]

        In the case of a real filter, only the nonzero region between ``[0, pi]``
        (half-spectrum) is returned. No wrapping can occur since it would inevitably
        interfere with itself due to conjugate symmetry. The half-spectrum can easily be
        recovered by:

        >>> half_width = (width + width % 2) // 2 + 1 - width % 2
        >>> half = numpy.zeros(half_width, dtype=trnc.dtype)
        >>> half[bin_idx:bin_idx + len(trnc)] = trnc

        And the full spectrum by:

        >>> full[bin_idx:bin_idx + len(trnc)] = trnc
        >>> full[width - bin_idx - len(trnc) + 1:width - bin_idx + 1] = \\
        ...     trnc[:None if bin_idx else 0:-1].conj()

        (the embedded if-statement is necessary when bin_idx is 0, as the full fft
        excludes its symmetric bin)

        Parameters
        ----------
        filt_idx
            The index of the filter to generate. Less than `num_filts`
        width
            The length of the DFT to output

        Returns
        -------
        tfr: tuple of int, array
        """
        pass


class TriangularOverlappingFilterBank(LinearFilterBank):
    """Triangular frequency response whose vertices are along the scale

    The vertices of the filters are sampled uniformly along the passed scale. If the
    scale is nonlinear, the triangles will be asymmetrical. This is closely related to,
    but not identical to, the filters described in [povey2011]_ and [young]_.

    Parameters
    ----------
    scaling_function
        Dictates the layout of filters in the Fourier domain. Can be a
        :class:`ScalingFunction` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    num_filts
        The number of filters in the bank
    high_hz
        The topmost edge of the filter frequencies. The default is the Nyquist for
        `sampling_rate`.
    low_hz
        The bottommost edge of the filter frequences.
    sampling_rate
        The sampling rate (cycles/sec) of the target recordings
    analytic
        Whether to use an analytic form of the bank. The analytic form is easily derived
        from the real form in [povey2011]_ and [young]_. Since the filter is compactly
        supported in frequency, the analytic form is simply the suppression of the
        ``[-pi, 0)`` frequencies

    Raises
    ------
    ValueError
        If `high_hz` is above the Nyquist, or `low_hz` is below :obj:`0`, or
        ``high_hz <= low_hz``
    """

    aliases = {"tri", "triangular"}  #:

    def __init__(
        self,
        scaling_function: Union[ScalingFunction, Mapping, str],
        num_filts: int = 40,
        high_hz: Optional[float] = None,
        low_hz: float = 20.0,
        sampling_rate: float = 16000,
        analytic: bool = False,
    ):
        scaling_function = alias_factory_subclass_from_arg(
            ScalingFunction, scaling_function
        )
        if low_hz < 0 or (
            high_hz and (high_hz <= low_hz or high_hz > sampling_rate // 2)
        ):
            raise ValueError(
                "Invalid frequency range: ({:.2f},{:.2f}".format(low_hz, high_hz)
            )
        self._rate = sampling_rate
        if high_hz is None:
            high_hz = sampling_rate // 2
        # compute vertices
        scale_low = scaling_function.hertz_to_scale(low_hz)
        scale_high = scaling_function.hertz_to_scale(high_hz)
        scale_delta = (scale_high - scale_low) / (num_filts + 1)
        self._vertices = tuple(
            scaling_function.scale_to_hertz(scale_low + scale_delta * idx)
            for idx in range(0, num_filts + 2)
        )
        self._analytic = analytic

    @property
    def is_real(self) -> bool:
        return not self._analytic

    @property
    def is_analytic(self) -> bool:
        return self._analytic

    @property
    def is_zero_phase(self) -> bool:
        return True

    @property
    def num_filts(self) -> int:
        return len(self._vertices) - 2

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def centers_hz(self) -> Tuple[float, ...]:
        """The point of maximum gain in each filter's frequency response, in Hz

        This property gives the so-called "center frequencies" - the point of maximum
        gain - of each filter.
        """
        return self._vertices[1:-1]

    @property
    def supports_hz(self) -> Tuple[Tuple[float, float], ...]:
        return tuple(
            (low, high) for low, high in zip(self._vertices[:-2], self._vertices[2:])
        )

    @property
    def supports(self) -> Tuple[Tuple[float, float], ...]:
        # A given filter is bound from above by
        # 2(w_r - w_l) / ((w_c - w_l)(w_r - w_c)t^2pi)
        supports = []
        for idx in range(len(self._vertices) - 2):
            left = hertz_to_angular(self._vertices[idx], self._rate)
            mid = hertz_to_angular(self._vertices[idx + 1], self._rate)
            right = hertz_to_angular(self._vertices[idx + 2], self._rate)
            K = np.sqrt(8 * (right - left) / np.pi)
            K /= np.sqrt(config.EFFECTIVE_SUPPORT_THRESHOLD)
            K /= np.sqrt(mid - left) * np.sqrt(right - mid)
            K = int(np.ceil(K))
            supports.append((-K // 2 - 1, K // 2 + 1))
        return tuple(supports)

    def get_impulse_response(self, filt_idx: int, width: int) -> np.ndarray:
        left = hertz_to_angular(self._vertices[filt_idx], self._rate)
        mid = hertz_to_angular(self._vertices[filt_idx + 1], self._rate)
        right = hertz_to_angular(self._vertices[filt_idx + 2], self._rate)
        res = np.zeros(width, dtype=np.complex128 if self._analytic else np.float64)
        # for numerical stability (angles can get pretty small)
        if right - mid > mid - left:
            denom = right - mid
            div_term = mid - left
        else:
            denom = mid - left
            div_term = right - mid
        denom *= (int(self._analytic) + 1) * np.pi
        for t in range(1, width + 1):
            if self._analytic:
                numer = (right - left) / div_term * np.exp(1j * mid * t)
                numer -= (right - mid) / div_term * np.exp(1j * left * t)
                numer -= (mid - left) / div_term * np.exp(1j * right * t)
            else:
                numer = (right - left) / div_term * np.cos(mid * t)
                numer -= (right - mid) / div_term * np.cos(left * t)
                numer -= (mid - left) / div_term * np.cos(right * t)
            val = numer / t ** 2
            if t < width:
                res[t] += val
                res[-t] += val.conj()
            else:
                res[0] += val
        numer = mid / div_term * (right ** 2 - left ** 2)
        numer += right / div_term * (left ** 2 - mid ** 2)
        numer += left / div_term * (mid ** 2 - right ** 2)
        res[0] += numer / 2
        res /= denom
        return res

    def get_frequency_response(
        self, filt_idx: int, width: int, half: bool = False
    ) -> np.ndarray:
        left = self._vertices[filt_idx]
        mid = self._vertices[filt_idx + 1]
        right = self._vertices[filt_idx + 2]
        left_idx = int(np.ceil(width * left / self._rate))
        right_idx = int(width * right / self._rate)
        assert self._rate * (left_idx - 1) / width <= left
        assert self._rate * (right_idx + 1) / width >= right, width
        dft_size = width
        if half:
            if width % 2:
                dft_size = (width + 1) // 2
            else:
                dft_size = width // 2 + 1
        res = np.zeros(dft_size, dtype=np.float64)
        for idx in range(left_idx, min(dft_size, right_idx + 1)):
            hz = self._rate * idx / width
            if hz <= mid:
                val = (hz - left) / (mid - left)
            else:
                val = (right - hz) / (right - mid)
            res[idx] = val
            if not half and not self._analytic:
                res[-idx] = val
        return res

    def get_truncated_response(
        self, filt_idx: int, width: int
    ) -> Tuple[int, np.ndarray]:
        left = self._vertices[filt_idx]
        mid = self._vertices[filt_idx + 1]
        right = self._vertices[filt_idx + 2]
        left_idx = int(np.ceil(width * left / self._rate))
        right_idx = int(width * right / self._rate)
        assert self._rate * (left_idx - 1) / width <= left
        assert self._rate * (right_idx + 1) / width >= right, width
        res = np.zeros(1 + right_idx - left_idx, dtype=np.float64)
        for idx in range(left_idx, min(width, right_idx + 1)):
            hz = self._rate * idx / width
            if hz <= mid:
                res[idx - left_idx] = (hz - left) / (mid - left)
            else:
                res[idx - left_idx] = (right - hz) / (right - mid)
        return left_idx, res


class Fbank(LinearFilterBank):
    """A mel-triangular filter bank that is square-rooted

    An ``Fbank`` instance is intended to replicate the filters from Kaldi [povey2011]_
    and HTK [young]_. Its scale is fixed to Mel-scale. Like a
    ``TriangularOverlappingFilterBank``, ``Fbank`` places the vertices of triangular
    filters uniformly along the target scale. However, an ``Fbank`` is triangular in the
    Mel-scale, whereas the triangular bank is triangular in frequency.

    Parameters
    ----------
    num_filts
        The number of filters in the bank
    high_hz
        The topmost edge of the filter frequencies. The default is the Nyquist for
        `sampling_rate`.
    low_hz
        The bottommost edge of the filter frequences.
    sampling_rate 
        The sampling rate (cycles/sec) of the target recordings
    analytic
        Whether to use an analytic form of the bank. The analytic form is easily derived
        from the real form in [povey2011]_ and [young]_. Since the filter is compactly
        supported in frequency, the analytic form is simply the suppression of the
        ``[-pi, 0)`` frequencies

    Notes
    -----
    In a standard mel-filterbank spectrogram, the power spectrum is calculated before
    filtering. This module's spectrogram takes the power spectrum after filtering. To
    recreate the frequency response of the alternate order, we can take the pointwise
    square root of the frequency response.
    """

    aliases = {"fbank"}  #:

    def __init__(
        self,
        num_filts: int = 40,
        high_hz: Optional[float] = None,
        low_hz: float = 20.0,
        sampling_rate: float = 16000,
        analytic: bool = False,
    ):
        scaling_function = MelScaling()
        if low_hz < 0 or (
            high_hz and (high_hz <= low_hz or high_hz > sampling_rate // 2)
        ):
            raise ValueError(
                "Invalid frequency range: ({:.2f},{:.2f}".format(low_hz, high_hz)
            )
        self._rate = sampling_rate
        if high_hz is None:
            high_hz = sampling_rate // 2
        # compute vertices
        scale_low = scaling_function.hertz_to_scale(low_hz)
        scale_high = scaling_function.hertz_to_scale(high_hz)
        scale_delta = (scale_high - scale_low) / (num_filts + 1)
        self._vertices = tuple(
            scaling_function.scale_to_hertz(scale_low + scale_delta * idx)
            for idx in range(0, num_filts + 2)
        )
        self._analytic = analytic

    @property
    def is_real(self) -> bool:
        return not self._analytic

    @property
    def is_analytic(self) -> bool:
        return self._analytic

    @property
    def is_zero_phase(self) -> bool:
        return True

    @property
    def num_filts(self) -> int:
        return len(self._vertices) - 2

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def centers_hz(self) -> Tuple[float, ...]:
        """tuple : The point of maximum gain in each filter's frequency response, in Hz

        This property gives the so-called "center frequencies" - the point of maximum
        gain - of each filter.
        """
        return self._vertices[1:-1]

    @property
    def supports_hz(self) -> Tuple[Tuple[float, float], ...]:
        return tuple(
            (low, high) for low, high in zip(self._vertices[:-2], self._vertices[2:])
        )

    @property
    def supports(self) -> Tuple[Tuple[float, float], ...]:
        # A given filter is bound above for t > 0 by
        # ((w_r - w_c) ** .5 + (w_c - w_l) ** .5) /
        #   (2 ** 3 * t ** 3 * (w_c - w_l) * (w_r - w_c) * pi) ** .5
        supports = []
        for idx in range(len(self._vertices) - 2):
            left = hertz_to_angular(self._vertices[idx], self._rate)
            mid = hertz_to_angular(self._vertices[idx + 1], self._rate)
            right = hertz_to_angular(self._vertices[idx + 2], self._rate)
            K = right - left + 2 * ((right - mid) * (mid - left)) ** 2
            K /= config.EFFECTIVE_SUPPORT_THRESHOLD ** 2 * np.pi
            K /= (right - mid) * (mid - left)
            K /= np.sqrt(config.EFFECTIVE_SUPPORT_THRESHOLD)
            K /= np.sqrt(mid - left) * np.sqrt(right - mid)
            K **= 0.3333
            K = int(np.ceil(K))
            supports.append((-K // 2 - 1, K // 2 + 1))
        return tuple(supports)

    def get_impulse_response(self, filt_idx: int, width: int) -> np.ndarray:
        # For the time being, I'll just invert the frequency response
        if self.is_analytic:
            freq_response = self.get_frequency_response(filt_idx, width, half=False)
            return np.fft.ifft(freq_response)
        else:
            freq_response = self.get_frequency_response(filt_idx, width, half=True)
            return np.fft.irfft(freq_response, n=width)

    def get_frequency_response(
        self, filt_idx: int, width: int, half: bool = False
    ) -> np.ndarray:
        scaling_function = MelScaling()
        left_hz = self._vertices[filt_idx]
        mid_hz = self._vertices[filt_idx + 1]
        right_hz = self._vertices[filt_idx + 2]
        left_mel = scaling_function.hertz_to_scale(left_hz)
        mid_mel = scaling_function.hertz_to_scale(mid_hz)
        right_mel = scaling_function.hertz_to_scale(right_hz)
        left_idx = int(np.ceil(width * left_hz / self._rate))
        right_idx = int(width * right_hz / self._rate)
        assert self._rate * (left_idx - 1) / width <= left_hz
        assert self._rate * (right_idx + 1) / width >= right_hz, width
        dft_size = width
        if half:
            if width % 2:
                dft_size = (width + 1) // 2
            else:
                dft_size = width // 2 + 1
        res = np.zeros(dft_size, dtype=np.float64)
        for idx in range(left_idx, min(dft_size, right_idx + 1)):
            hz = self._rate * idx / width
            mel = scaling_function.hertz_to_scale(hz)
            if mel <= mid_mel:
                val = (mel - left_mel) / (mid_mel - left_mel)
            else:
                val = (right_mel - mel) / (right_mel - mid_mel)
            res[idx] = val ** 0.5
            if not half and not self._analytic:
                res[-idx] = val ** 0.5
        return res

    def get_truncated_response(
        self, filt_idx: int, width: int
    ) -> Tuple[int, np.ndarray]:
        scaling_function = MelScaling()
        left_hz = self._vertices[filt_idx]
        mid_hz = self._vertices[filt_idx + 1]
        right_hz = self._vertices[filt_idx + 2]
        left_mel = scaling_function.hertz_to_scale(left_hz)
        mid_mel = scaling_function.hertz_to_scale(mid_hz)
        right_mel = scaling_function.hertz_to_scale(right_hz)
        left_idx = int(np.ceil(width * left_hz / self._rate))
        right_idx = int(width * right_hz / self._rate)
        assert self._rate * (left_idx - 1) / width <= left_hz
        assert self._rate * (right_idx + 1) / width >= right_hz, width
        res = np.zeros(min(width, right_idx + 1) - left_idx, dtype=np.float64)
        for idx in range(left_idx, min(width, right_idx + 1)):
            hz = self._rate * idx / width
            mel = scaling_function.hertz_to_scale(hz)
            if mel <= mid_mel:
                res[idx - left_idx] = (mel - left_mel) / (mid_mel - left_mel)
            else:
                res[idx - left_idx] = (right_mel - mel) / (right_mel - mid_mel)
        return left_idx, res ** 0.5


class GaborFilterBank(LinearFilterBank):
    r"""Gabor filters with ERBs between points from a scale

    Gabor filters are complex, mostly analytic filters that have a Gaussian envelope in
    both the time and frequency domains. They are defined as

    .. math::

         f(t) = C \sigma^{-1/2} \pi^{-1/4}
                e^{\frac{-t^2}{2\sigma^2} + i\xi t}

    in the time domain and

    .. math::

         \widehat{f}(\omega) = C \sqrt{2\sigma} \pi^{1/4}
                               e^{\frac{-\sigma^2(\xi - \omega)^2}{2}}

    in the frequency domain. Though Gaussians never truly reach 0, in either domain,
    they are effectively compactly supported. Gabor filters are optimal with respect to
    their time-bandwidth product.

    `scaling_function` is used to split up the frequencies between `high_hz` and
    `low_hz` into a series of filters. Every subsequent filter's width is scaled such
    that, if the filters are all of the same height, the intersection with the precedent
    filter's response matches the filter's Equivalent Rectangular Bandwidth (``erb ==
    True``) or its 3dB bandwidths (``erb == False``). The ERB is the width of a
    rectangular filter with the same height as the filter's maximum frequency response
    that has the same :math:`L^2` norm.

    Parameters
    ----------
    scaling_function
        Dictates the layout of filters in the Fourier domain. Can be a
        :class:`ScalingFunction` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    num_filts
        The number of filters in the bank
    high_hz
        The topmost edge of the filter frequencies. The default is the Nyquist for
        `sampling_rate`.
    low_hz
        The bottommost edge of the filter frequences.
    sampling_rate
        The sampling rate (cycles/sec) of the target recordings
    scale_l2_norm
        Whether to scale the l2 norm of each filter to 1. Otherwise the frequency
        response of each filter will max out at an absolute value of 1.
    erb : bool

    See Also
    --------
    pydrobert.speech.config.EFFECTIVE_SUPPORT_THRESHOLD
        The absolute value below which counts as zero
    """

    aliases = {"gabor"}  #:

    def __init__(
        self,
        scaling_function: Union[ScalingFunction, Mapping, str],
        num_filts: int = 40,
        high_hz: Optional[float] = None,
        low_hz: float = 20.0,
        sampling_rate: float = 16000,
        scale_l2_norm: bool = False,
        erb: bool = False,
    ):
        scaling_function = alias_factory_subclass_from_arg(
            ScalingFunction, scaling_function
        )
        self._scale_l2_norm = scale_l2_norm
        self._erb = erb
        if low_hz < 0 or (
            high_hz and (high_hz <= low_hz or high_hz > sampling_rate // 2)
        ):
            raise ValueError(
                "Invalid frequency range: ({:.2f},{:.2f}".format(low_hz, high_hz)
            )
        self._rate = sampling_rate
        if high_hz is None:
            high_hz = sampling_rate // 2
        scale_low = scaling_function.hertz_to_scale(low_hz)
        scale_high = scaling_function.hertz_to_scale(high_hz)
        scale_delta = (scale_high - scale_low) / (num_filts + 1)
        # edges dictate the points where filters should intersect. We
        # make a pretend intersection halfway between low_hz and
        # the first filter center in the scaled domain. Likewise with
        # high_hz and the last filter center. Intersections are spaced
        # uniformly in the scaled domain
        edges = tuple(
            scaling_function.scale_to_hertz(scale_low + scale_delta * (idx + 0.5))
            for idx in range(0, num_filts + 1)
        )
        centers_hz = []
        centers_ang = []
        stds = []
        supports_ang = []
        supports = []
        wrap_supports_ang = []
        self._wrap_below = False
        log_2 = np.log(2)
        log_pi = np.log(np.pi)
        t_support_const = -2 * np.log(config.EFFECTIVE_SUPPORT_THRESHOLD)
        f_support_const = t_support_const
        if scale_l2_norm:
            f_support_const += log_2 + 0.5 * log_pi
            t_support_const -= 0.5 * log_pi
        else:
            t_support_const -= log_2 + log_pi
        if erb:
            bandwidth_const = np.sqrt(np.pi) / 2
        else:
            bandwidth_const = np.sqrt(3 / 10 * np.log(10))
        for left_intersect, right_intersect in zip(edges[:-1], edges[1:]):
            center_hz = (left_intersect + right_intersect) / 2
            center_ang = hertz_to_angular(center_hz, self._rate)
            std = bandwidth_const / hertz_to_angular(
                center_hz - left_intersect, self._rate
            )
            log_std = np.log(std)
            if scale_l2_norm:
                diff_ang = np.sqrt(log_std + f_support_const) / std
                wrap_diff_ang = np.sqrt(log_std + f_support_const + log_2) / std
                diff_samps = int(np.ceil(std * np.sqrt(t_support_const - log_std)))
            else:
                diff_ang = np.sqrt(f_support_const) / std
                wrap_diff_ang = np.sqrt(f_support_const + log_2) / std
                diff_samps = int(np.ceil(std * np.sqrt(t_support_const - 2 * log_std)))
            supp_ang_low = center_ang - diff_ang
            if supp_ang_low < 0:
                self._wrap_below = True
            centers_hz.append(center_hz)
            centers_ang.append(center_ang)
            supports_ang.append((center_ang - diff_ang, center_ang + diff_ang))
            wrap_supports_ang.append(2 * wrap_diff_ang)
            supports.append((-diff_samps, diff_samps))
            stds.append(std)
        self._centers_ang = tuple(centers_ang)
        self._centers_hz = tuple(centers_hz)
        self._stds = tuple(stds)
        self._supports_ang = tuple(supports_ang)
        self._wrap_supports_ang = tuple(wrap_supports_ang)
        self._supports_hz = tuple(
            (angular_to_hertz(ang_l, self._rate), angular_to_hertz(ang_h, self._rate),)
            for ang_l, ang_h in supports_ang
        )
        self._supports = tuple(supports)

    @property
    def is_real(self) -> bool:
        return False

    @property
    def is_analytic(self) -> bool:
        return not self._wrap_below

    @property
    def num_filts(self) -> int:
        return len(self._centers_hz)

    @property
    def is_zero_phase(self) -> bool:
        return True

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def centers_hz(self) -> Tuple[float, ...]:
        """tuple :The point of maximum gain in each filter's frequency response, in Hz

        This property gives the so-called "center frequencies" - the
        point of maximum gain - of each filter.
        """
        return self._centers_hz

    @property
    def supports_hz(self) -> Tuple[Tuple[float, float], ...]:
        return self._supports_hz

    @property
    def supports(self) -> Tuple[Tuple[float, float], ...]:
        return self._supports

    @property
    def scaled_l2_norm(self) -> bool:
        return self._scale_l2_norm

    @property
    def erb(self) -> bool:
        return self._erb

    def get_impulse_response(self, filt_idx: int, width: int) -> np.ndarray:
        center_ang = self._centers_ang[filt_idx]
        std = self._stds[filt_idx]
        res = np.zeros(width, dtype=np.complex128)
        if self._scale_l2_norm:
            const_term = -0.5 * np.log(std) - 0.25 * np.log(np.pi)
        else:
            const_term = -0.5 * np.log(2 * np.pi) - np.log(std)
        denom_term = 2 * std ** 2
        for t in range(width + 1):
            val = -(t ** 2) / denom_term + const_term + 1j * center_ang * t
            val = np.exp(val)
            if t != width:
                res[t] += val
            if t:
                res[-t] += val.conj()
        return res

    def get_frequency_response(
        self, filt_idx: int, width: int, half: bool = False
    ) -> np.ndarray:
        center_ang = self._centers_ang[filt_idx]
        lowest_ang, highest_ang = self._supports_ang[filt_idx]
        std = self._stds[filt_idx]
        dft_size = width
        if half:
            if width % 2:
                dft_size = (width + 1) // 2
            else:
                dft_size = width // 2 + 1
        res = np.zeros(dft_size, dtype=np.float64)
        if self._scale_l2_norm:
            const_term = 0.5 * np.log(2 * std) + 0.25 * np.log(np.pi)
        else:
            const_term = 0
        num_term = -(std ** 2) / 2
        for idx in range(dft_size):
            for period in range(
                -1 - int(max(-lowest_ang, 0) / (2 * np.pi)),
                2 + int(highest_ang / (2 * np.pi)),
            ):
                omega = (idx / width + period) * 2 * np.pi
                val = num_term * (center_ang - omega) ** 2 + const_term
                val = np.exp(val)
                res[idx] += val
        return res

    def get_truncated_response(
        self, filt_idx: int, width: int
    ) -> Tuple[int, np.ndarray]:
        # wrap_supports_ang contains the angular supports of each filter
        # if the effective support threshold were halved. If this
        # support exceeds the 2pi period, overlap from aliasing in the
        # periphery will exceed the effective support, meaning the
        # entire period lies in the support
        if self._wrap_supports_ang[filt_idx] >= 2 * np.pi:
            return 0, self.get_frequency_response(filt_idx, width)
        center_ang = self._centers_ang[filt_idx]
        std = self._stds[filt_idx]
        lowest_ang, highest_ang = self._supports_ang[filt_idx]
        left_idx = int(np.ceil(width * lowest_ang / (2 * np.pi)))
        right_idx = int(width * highest_ang / (2 * np.pi))
        res = np.zeros(1 + right_idx - left_idx, dtype=np.float64)
        if self._scale_l2_norm:
            const_term = 0.5 * np.log(2 * std) + 0.25 * np.log(np.pi)
        else:
            const_term = 0
        num_term = -(std ** 2) / 2
        for idx in range(left_idx, right_idx + 1):
            for period in range(
                -int(max(-lowest_ang, 0) / (2 * np.pi)),
                1 + int(highest_ang / (2 * np.pi)),
            ):
                omega = (idx / width + period) * 2 * np.pi
                val = num_term * (center_ang - omega) ** 2 + const_term
                val = np.exp(val)
                res[idx - left_idx] += val
        return left_idx % width, res


class ComplexGammatoneFilterBank(LinearFilterBank):
    r"""Gammatone filters with complex carriers

    A complex gammatone filter [flanagan1960]_ [aertsen1981]_ can be defined as

    .. math::

        h(t) = c t^{n - 1} e^{- \alpha t + i\xi t} u(t)

    in the time domain, where :math:`\alpha` is the bandwidth parameter, :math:`\xi` is
    the carrier frequency, :math:`n` is the order of the function, :math:`u(t)` is the
    step function, and :math:`c` is a normalization constant. In the frequency domain,
    the filter is defined as

    .. math::

        H(\omega) = \frac{c(n - 1)!}{\left(
            \alpha + i(\omega - \xi) \right)^n}

    For large :math:`\xi`, the complex gammatone is approximately analytic.

    `scaling_function` is used to split up the frequencies between `high_hz` and
    `low_hz` into a series of filters. Every subsequent filter's width is scaled such
    that, if the filters are all of the same height, the intersection with the precedent
    filter's response matches the filter's Equivalent Rectangular Bandwidth (``erb ==
    True``) or its 3dB bandwidths (``erb == False``). The ERB is the width of a
    rectangular filter with the same height as the filter's maximum frequency response
    that has the same :math:`L^2` norm.

    Parameters
    ----------
    scaling_function
        Dictates the layout of filters in the Fourier domain. Can be a
        :class:`ScalingFunction` or something compatible with
        :func:`pydrobert.speech.alias.alias_factory_subclass_from_arg`
    num_filts
        The number of filters in the bank
    high_hz
        The topmost edge of the filter frequencies. The default is the Nyquist for
        `sampling_rate`.
    low_hz
        The bottommost edge of the filter frequences.
    sampling_rate : float, optional
        The sampling rate (cycles/sec) of the target recordings
    order
        The :math:`n` parameter in the Gammatone. Should be positive. Larger orders
        will make the gammatone more symmetrical.
    max_centered
        While normally causal, setting `max_centered` to true will shift all filters in
        the bank such that the maximum absolute value in time is centered at sample 0.
    scale_l2_norm
        Whether to scale the l2 norm of each filter to 1. Otherwise the frequency
        response of each filter will max out at an absolute value of 1.
    erb

    See Also
    --------
    pydrobert.speech.config.EFFECTIVE_SUPPORT_THRESHOLD
        The absolute value below which counts as zero
    """

    aliases = {"gammatone", "tonebank"}  #:

    def __init__(
        self,
        scaling_function: Union[ScalingFunction, Mapping, str],
        num_filts: int = 40,
        high_hz: Optional[float] = None,
        low_hz: float = 20.0,
        sampling_rate: float = 16000,
        order: int = 4,
        max_centered: bool = False,
        scale_l2_norm: bool = False,
        erb: bool = False,
    ):
        scaling_function = alias_factory_subclass_from_arg(
            ScalingFunction, scaling_function
        )
        self._scale_l2_norm = scale_l2_norm
        self._erb = erb
        if low_hz < 0 or (
            high_hz and (high_hz <= low_hz or high_hz > sampling_rate // 2)
        ):
            raise ValueError(
                "Invalid frequency range: ({:.2f},{:.2f}".format(low_hz, high_hz)
            )
        if not isinstance(order, int) or order <= 0:
            raise ValueError("order must be a positive integer")
        self._order = order
        self._rate = sampling_rate
        if high_hz is None:
            high_hz = sampling_rate // 2
        scale_low = scaling_function.hertz_to_scale(low_hz)
        scale_high = scaling_function.hertz_to_scale(high_hz)
        scale_delta = (scale_high - scale_low) / (num_filts + 1)
        # see gabor filters for more info
        edges = tuple(
            scaling_function.scale_to_hertz(scale_low + scale_delta * (idx + 0.5))
            for idx in range(0, num_filts + 1)
        )
        self._centers_hz = []
        self._xis = []
        self._alphas = []
        self._cs = []
        self._offsets = []
        self._supports = []
        self._supports_ang = []
        self._wrap_supports_ang = []
        self._wrap_below = False
        log_eps = np.log(config.EFFECTIVE_SUPPORT_THRESHOLD)
        log_double_factorial = np.log(np.math.factorial(2 * order - 2))
        log_factorial = np.log(np.math.factorial(order - 1))
        log_2 = np.log(2)
        if erb:
            alpha_const = log_2 * (2 * order - 1)
            alpha_const += 2 * log_factorial
            alpha_const -= log_double_factorial
        else:
            alpha_const = -0.5 * np.log(4 * (2 ** (1 / order)) - 4)
        for left_intersect, right_intersect in zip(edges[:-1], edges[1:]):
            center_hz = (left_intersect + right_intersect) / 2
            xi = hertz_to_angular(center_hz, self._rate)
            log_alpha = alpha_const + np.log(
                hertz_to_angular(right_intersect - left_intersect, self._rate)
            )
            alpha = np.exp(log_alpha)
            if scale_l2_norm:
                log_c = 0.5 * (log_2 + log_alpha + log_double_factorial)
                log_c -= order * (log_alpha + log_2)
            else:
                log_c = order * log_alpha - log_factorial
            c = np.exp(log_c)
            if max_centered:
                offset = -(order - 1) / alpha
            else:
                offset = 0
            supp_a = (2 / order) * (log_c + log_factorial - log_eps)
            wrap_supp_a = supp_a + (2 / order) * log_2
            supp_b = np.exp(2 * log_alpha)
            diff_ang = (np.exp(supp_a) - supp_b) ** 0.5
            wrap_diff_ang = (np.exp(wrap_supp_a) - supp_b) ** 0.5
            self._centers_hz.append(center_hz)
            self._xis.append(xi)
            self._alphas.append(alpha)
            self._cs.append(c)
            self._offsets.append(offset)
            self._supports.append(self._calculate_temp_support(-1))
            self._supports_ang.append((xi - diff_ang, xi + diff_ang))
            if self._supports_ang[-1][0] < 0:
                self._wrap_below = True
            self._wrap_supports_ang.append(2 * wrap_diff_ang)
        self._xis = tuple(self._xis)
        self._cs = tuple(self._cs)
        self._alphas = tuple(self._alphas)
        self._offsets = tuple(self._offsets)
        self._centers_hz = tuple(self._centers_hz)
        self._supports_ang = tuple(self._supports_ang)
        self._wrap_supports_ang = tuple(self._wrap_supports_ang)
        self._supports_hz = tuple(
            (angular_to_hertz(ang_l, self._rate), angular_to_hertz(ang_h, self._rate),)
            for ang_l, ang_h in self._supports_ang
        )
        self._supports = tuple(self._supports)

    @property
    def is_real(self) -> bool:
        return False

    @property
    def is_analytic(self) -> bool:
        return not self._wrap_below

    @property
    def num_filts(self) -> int:
        return len(self._centers_hz)

    @property
    def order(self) -> int:
        return self._order

    @property
    def is_zero_phase(self) -> bool:
        return False

    @property
    def sampling_rate(self) -> float:
        return self._rate

    @property
    def centers_hz(self) -> Tuple[float, ...]:
        """int : The point of maximum gain in each filter's frequency response, in Hz

        This property gives the so-called "center frequencies" - the point of maximum
        gain - of each filter.
        """
        return self._centers_hz

    @property
    def supports_hz(self) -> Tuple[Tuple[float, float], ...]:
        return self._supports_hz

    @property
    def supports(self) -> Tuple[Tuple[float, float], ...]:
        return self._supports

    @property
    def scaled_l2_norm(self) -> bool:
        return self._scale_l2_norm

    @property
    def erb(self) -> bool:
        return self._erb

    def get_impulse_response(self, filt_idx: int, width: int) -> np.ndarray:
        left_sup, right_sup = self.supports[filt_idx]
        left_period = int(np.floor(left_sup / width))
        right_period = int(np.ceil(right_sup / width))
        res = np.zeros(width, dtype=np.complex128)
        for period in range(left_period, right_period + 1):
            for idx in range(width):
                t = period * width + idx
                res[idx] += self._h(t, filt_idx)
        return res

    def get_frequency_response(
        self, filt_idx: int, width: int, half: bool = False
    ) -> np.ndarray:
        left_sup, right_sup = self._supports_ang[filt_idx]
        left_period = int(np.floor(left_sup / 2 / np.pi))
        right_period = int(np.ceil(right_sup / 2 / np.pi))
        if half:
            if width % 2:
                dft_size = (width + 1) // 2
            else:
                dft_size = width // 2 + 1
        else:
            dft_size = width
        res = np.zeros(dft_size, dtype=np.complex128)
        omega = np.arange(dft_size, dtype=np.float64) * 2 * np.pi / width
        for period in range(left_period, right_period + 1):
            res += self._H(omega + 2 * np.pi * period, filt_idx)
        return res

    def get_truncated_response(
        self, filt_idx: int, width: int
    ) -> Tuple[int, np.ndarray]:
        left_sup, right_sup = self._supports_ang[filt_idx]
        wrap_ang = self._wrap_supports_ang[filt_idx]
        # wrap_ang is the additional support needed to hit
        # half the effective support threshold. If that support is
        # greater than the 2pi periodization, some points could exceed
        # the threshold due to wrapping.
        if right_sup - left_sup + wrap_ang >= 2 * np.pi:
            return 0, self.get_frequency_response(filt_idx, width)
        left_idx = int(np.ceil(width * left_sup / (2 * np.pi)))
        right_idx = int(width * right_sup / (2 * np.pi))
        omega = np.arange(left_idx, right_idx + 1, dtype=np.float64)
        omega *= 2 * np.pi / width
        return left_idx % width, self._H(omega, filt_idx)

    def _h(self, t, idx):
        # calculate impulse response of filt idx at sample t
        offset = self._offsets[idx]
        if t <= offset:
            return 0j
        alpha = self._alphas[idx]
        log_c = np.log(self._cs[idx])
        xi = self._xis[idx]
        n = self._order
        r = log_c + (n - 1) * np.log(t - offset)
        r += (-alpha + 1j * xi) * (t - offset)
        return np.exp(r)

    def _H(self, omega, idx):
        # calculate frequency response of filt idx at ang freqs omega
        alpha = self._alphas[idx]
        c = self._cs[idx]
        xi = self._xis[idx]
        offset = self._offsets[idx]
        n = self._order
        numer = np.exp(-1j * omega * offset) * c * np.math.factorial(n - 1)
        denom = (alpha + 1j * (omega - xi)) ** n
        return numer / denom

    def _calculate_temp_support(self, idx):
        # calculate the nonzero region of the temp support of filt idx
        alpha = self._alphas[idx]
        c = self._cs[idx]
        # xi = self._xis[idx]
        offset = self._offsets[idx]
        n = self._order
        eps = config.EFFECTIVE_SUPPORT_THRESHOLD
        if n == 1:
            right = int(np.ceil((np.log(c) - np.log(eps) / alpha)))
        else:

            def _d(t):
                # derivative of abs func
                v = c * np.exp(-alpha * t) * t ** (n - 2)
                v *= (n - 1) - alpha * t
                return v

            right = (n - 1 + np.sqrt((n - 1) / 2)) / alpha
            h_0 = np.abs(self._h(right, idx))
            while h_0 > eps:
                d_0 = _d(right)
                right -= h_0 / d_0
                h_0 = np.abs(self._h(right, idx))
        return (int(np.floor(offset)), int(np.ceil(right) + offset))


# windows


class WindowFunction(AliasedFactory):
    """A real linear filter, usually lowpass"""

    @abc.abstractmethod
    def get_impulse_response(self, width: int) -> np.ndarray:
        """Write the filter into a numpy array of fixed width
        
        Parameters
        ----------
        width
            The length of the window in samples
        
        Returns
        -------
        ir : np.ndarray
            A 1D vector of length `width`
        """
        pass


class BartlettWindow(WindowFunction):
    """A unit-normalized triangular window

    See Also
    --------
    numpy.bartlett
    """

    aliases = {"bartlett", "triangular", "tri"}  #:

    def get_impulse_response(self, width: int) -> np.ndarray:
        window = np.bartlett(width)
        window /= max(1, width - 1) / 2
        return window


class BlackmanWindow(WindowFunction):
    """A unit-normalized Blackman window

    See Also
    --------
    numpy.blackman
    """

    aliases = {"blackman", "black"}  #:

    def get_impulse_response(self, width: int) -> np.ndarray:
        window = np.blackman(width)
        window /= 0.42 * max(1, width - 1)
        return window


class HammingWindow(WindowFunction):
    """A unit-normalized Hamming window

    See Also
    --------
    numpy.hamming
    """

    aliases = {"hamming"}  #:

    def get_impulse_response(self, width: int) -> np.ndarray:
        window = np.hamming(width)
        window /= 0.54 * max(1, width - 1)
        return window


class HannWindow(WindowFunction):
    """A unit-normalized Hann window

    See Also
    --------
    numpy.hanning
    """

    aliases = {"hanning", "hann"}  #:

    def get_impulse_response(self, width: int) -> np.ndarray:
        window = np.hanning(width)
        window /= 0.5 * max(1, width - 1)
        return window


class GammaWindow(WindowFunction):
    r"""A lowpass filter based on the Gamma function

    A Gamma function is defined as:

    .. math:: p(t; \alpha, n) = t^{n - 1} e^{-\alpha t} u(t)

    Where :math:`n` is the order of the function, :math:`\alpha` controls the bandwidth
    of the filter, and :math:`u` is the step function.

    This function returns a window based off a reflected Gamma function. :math:`\alpha`
    is chosen such that the maximum value of the window aligns with `peak`. The window
    is clipped to the width. For reasonable values of `peak` (i.e. in the last quarter
    of samples), the majority of the support should lie in this interval anyways.

    Arguments
    ---------
    order
    peak
        ``peak * width``, where ``width`` is the length of the window in samples, is
        where the approximate maximal value of the window lies
    """

    order: int  #:
    peak: float  #:
    aliases = {"gamma"}  #:

    def __init__(self, order: int = 4, peak: float = 0.75):
        self.order = order
        self.peak = peak

    def get_impulse_response(self, width: np.ndarray) -> np.ndarray:
        if width <= 0:
            return np.array([], dtype=float)
        elif width == 1:
            return np.array([1], dtype=float)
        peak = self.peak * width
        ret = np.arange(width - 1, -1, -1, dtype=float)
        if self.order > 1:
            alpha = (self.order - 1) / (width - peak)
            offs = width - 1
        else:
            # align alpha roughly with a support in M
            alpha = 5 / width
            offs = width
        ln_c = self.order * np.log(alpha)
        ln_c -= np.log(np.math.factorial(self.order - 1))
        ret[:offs] = ret[:offs] ** (self.order - 1) * np.exp(-alpha * ret[:offs] + ln_c)
        return ret



# Copyright 2021 Sean Robertson

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Script-like functions intended to be accessed by command line"""


import os
import argparse
import json
import logging
import sys
from typing import Optional, Sequence

import numpy as np

import pydrobert.speech as speech
import pydrobert.speech.config as config

from pydrobert.speech.compute import FrameComputer
from pydrobert.speech.pre import PreProcessor
from pydrobert.speech.post import PostProcessor
from pydrobert.speech.alias import alias_factory_subclass_from_arg
from pydrobert.speech.util import read_signal

try:
    from pydrobert.kaldi.logging import kaldi_vlog_level_cmd_decorator  # type: ignore
    from pydrobert.kaldi.logging import kaldi_logger_decorator  # type: ignore
except ImportError:

    def kaldi_vlog_level_cmd_decorator(func):
        return func

    def kaldi_logger_decorator(func):
        return func


try:
    import torch.utils.data

    class _FeatureProcessorDataset(torch.utils.data.Dataset):
        # yields utt, feats

        def __init__(
            self,
            utt2path,
            preprocessors,
            computer,
            postprocessors,
            channel,
            force_as,
            seed,
        ):
            super(_FeatureProcessorDataset, self).__init__()
            self.utt_path = tuple(utt2path.items())
            self.preprocessors = preprocessors
            self.computer = computer
            self.postprocessors = postprocessors
            self.channel = channel
            self.force_as = force_as
            self.seed = seed

        def __len__(self):
            return len(self.utt_path)

        def __getitem__(self, idx):
            # torch.multiprocessing should copy numpy rng over workers, so
            # we need not worry about the rng changing underneath us
            np.random.seed(self.seed + idx)
            utt_id, path = self.utt_path[idx]
            try:
                signal = read_signal(
                    path, dtype=np.float64, force_as=self.force_as, key=utt_id
                )
            except Exception as e:
                raise IOError("Utterance {}:".format(utt_id), e) from e
            if self.channel == -1 and len(signal.shape) > 1 and signal.shape[0] > 1:
                raise ValueError(
                    "Utterance {}: Channel is not specified but signal has "
                    "shape {}".format(utt_id, signal.shape)
                )
            elif (self.channel != -1 and len(signal.shape) == 1) or (
                self.channel >= signal.shape[0]
            ):
                raise ValueError(
                    "Utterance {}: Channel specified as {} but signal has "
                    "shape {}".format(utt_id, self.channel, signal.shape)
                )
            if len(signal.shape) != 1:
                signal = signal[self.channel]
            for preprocessor in self.preprocessors:
                signal = preprocessor.apply(signal, in_place=True)
            if self.computer is None:
                feats = signal[:, None]  # add "filter" dim to raw signal
            else:
                feats = self.computer.compute_full(signal)
            del signal
            for postprocessor in self.postprocessors:
                feats = postprocessor.apply(feats, in_place=True)
            feats = torch.FloatTensor(feats)
            return utt_id, feats


except ImportError:
    pass

__all__ = [
    "compute_feats_from_kaldi_tables",
    "signals_to_torch_feat_dir",
]


def _json_type(string):
    """Convert JSON string (or path to JSON file) to container hierarchy"""
    try:
        with open(string) as file_obj:
            return json.load(file_obj)
    except ValueError:
        raise argparse.ArgumentTypeError(
            'Unable to parse file as json: "{}"'.format(string)
        )
    except IOError:
        pass
    try:
        return json.loads(string)
    except ValueError:
        raise argparse.ArgumentTypeError(
            'Unable to parse string as json: "{}"'.format(string)
        )


def _nonneg_int_type(string):
    """Convert to an int and make sure its nonnegative"""
    try:
        val = int(string)
        assert val >= 0
    except (ValueError, AssertionError):
        raise argparse.ArgumentTypeError(
            "{} is not a nonnegative integer".format(string)
        )
    return val


def _compute_feats_from_kaldi_tables_parse_args(args, logger):
    from pydrobert.kaldi.io.argparse import KaldiParser  # type: ignore

    parser = KaldiParser(
        description=compute_feats_from_kaldi_tables.__doc__,
        add_verbose=True,
        logger=logger,
        version=speech.__version__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "wav_rspecifier", type="kaldi_rspecifier", help="Input wave table rspecifier"
    )
    parser.add_argument(
        "feats_wspecifier",
        type="kaldi_wspecifier",
        help="Output feature table wspecifier",
    )
    parser.add_argument(
        "computer_config",
        type=_json_type,
        help="JSON file or string to configure a "
        "'pydrobert.speech.compute.FrameComputer' object to calculate "
        "features with",
    )
    parser.add_argument(
        "--min-duration",
        type=float,
        default=0.0,
        help="Min duration of segments to process (in seconds)",
    )
    parser.add_argument(
        "--channel",
        type=int,
        default=-1,
        help="Channel to draw audio from. Default is to assume mono",
    )
    parser.add_argument(
        "--preprocess",
        type=_json_type,
        default=tuple(),
        help="JSON list of configurations for "
        "'pydrobert.speech.pre.PreProcessor' objects. Audio will be "
        "preprocessed in the same order as the list",
    )
    parser.add_argument(
        "--postprocess",
        type=_json_type,
        default=tuple(),
        help="JSON List of configurations for "
        "'pydrobert.speech.post.PostProcessor' objects. Features will be "
        "postprocessed in the same order as the list",
    )
    parser.add_argument(
        "--seed",
        type=_nonneg_int_type,
        default=None,
        help="A random seed used for determinism. This affects operations "
        "like dithering. If unset, a seed will be generated at the moment",
    )
    return parser.parse_args(args)


@kaldi_vlog_level_cmd_decorator
@kaldi_logger_decorator
def compute_feats_from_kaldi_tables(args: Optional[Sequence[str]] = None) -> None:
    """Store features from a kaldi archive in a kaldi archive

    This command is intended to replace Kaldi's (https://kaldi-asr.org/) series of
    "compute-<something>-feats" scripts in a Kaldi pipeline.
    """
    from pydrobert.kaldi.logging import register_logger_for_kaldi  # type: ignore
    from pydrobert.kaldi.io.enums import KaldiDataType  # type: ignore
    from pydrobert.kaldi.io import open as kaldi_open  # type: ignore

    logger = logging.getLogger(sys.argv[0])
    logger.addHandler(logging.StreamHandler())
    register_logger_for_kaldi(logger)
    try:
        options = _compute_feats_from_kaldi_tables_parse_args(args, logger)
    except SystemExit as ex:
        return ex.code
    if options.seed is not None:
        np.random.seed(options.seed)
    # construct the computer
    try:
        computer = alias_factory_subclass_from_arg(
            FrameComputer, options.computer_config
        )
    except ValueError:
        logger.error("Failed to build computer:", exc_info=True)
        return 1
    # construct the preprocessors (if any)
    preprocessors = []
    try:
        if isinstance(options.preprocess, dict):
            preprocessors.append(
                alias_factory_subclass_from_arg(PreProcessor, options.preprocess)
            )
        else:
            for element in options.preprocess:
                preprocessors.append(
                    alias_factory_subclass_from_arg(PreProcessor, element)
                )
    except ValueError:
        logger.error("Failed to build preprocessor:", exc_info=True)
        return 1
    postprocessors = []
    try:
        if isinstance(options.postprocess, dict):
            postprocessors.append(
                alias_factory_subclass_from_arg(PostProcessor, options.postprocess)
            )
        else:
            for element in options.postprocess:
                postprocessors.append(
                    alias_factory_subclass_from_arg(PostProcessor, element)
                )
    except ValueError:
        logger.error("Failed to build postprocessor:", exc_info=True)
        return 1
    # open tables
    try:
        wav_reader = kaldi_open(options.wav_rspecifier, "wm", value_style="bsd")
    except IOError:
        logger.error("Could not read the wave table {}".format(options.wav_rspecifier))
        return 1
    try:
        feat_writer = kaldi_open(options.feats_wspecifier, "bm", mode="w")
    except IOError:
        logger.error(
            "Could not open the feat table {} for writing".format(
                options.feats_wspecifier
            )
        )
        return 1
    num_utts, num_success = 0, 0
    for utt_id, (buff, samp_freq, duration) in list(wav_reader.items()):
        num_utts += 1
        if duration < options.min_duration:
            logger.warn(
                "File: {} is too short ({:.2f} sec): producing no output"
                "".format(utt_id, duration)
            )
            continue
        elif samp_freq != computer.bank.sampling_rate:
            logger.warn(
                "Sample frequency mismatch for file {}: you specified {:.2f} "
                "but data has {:.2f}: producing no output"
                "".format(utt_id, computer.bank.sample_rate_hz, samp_freq)
            )
            continue
        cur_chan = options.channel
        if options.channel == -1 and buff.shape[0] > 1:
            logger.warning(
                "Channel is not specified but you have data with {} channels;"
                " defaulting to zero".format(buff.shape[0])
            )
            cur_chan = 0
        elif options.channel >= buff.shape[0]:
            logger.warn(
                "File with id {} has {} channels but you specified channel {},"
                " producing no ouptut".format(utt_id, buff.shape[0], options.channel)
            )
            continue
        buff = buff[cur_chan].astype(np.float64, copy=False)
        for preprocessor in preprocessors:
            buff = preprocessor.apply(buff, in_place=True)
        feats = computer.compute_full(buff)
        if not KaldiDataType.BaseMatrix.is_double:
            feats = feats.astype(np.float32)
        feat_writer.write(utt_id, feats)
        if num_utts % 10 == 0:
            logger.info("Processed {} utterances".format(num_utts))
        logger.log(9, "Processed features for key {}".format(utt_id))
        num_success += 1
    logger.info("Done {} out of {} utterances".format(num_success, num_utts))
    feat_writer.close()
    wav_reader.close()
    return 0 if num_success else 1


def _signals_to_torch_feat_dir_parse_args(args):
    parser = argparse.ArgumentParser(
        description=signals_to_torch_feat_dir.__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "map",
        type=argparse.FileType("r"),
        help="Path to the file containing (<utterance>, <path>) pairs",
    )
    parser.add_argument(
        "computer_config",
        type=_json_type,
        nargs="?",
        default=None,
        help="JSON file or string to configure a "
        "pydrobert.speech.compute.FrameComputer object to calculate features with. If "
        "unspecified, the audio (with channels removed)  will be stored directly with "
        "shape (S, 1), where S is the number of samples",
    )
    parser.add_argument(
        "dir",
        help="Directory to output features to. If the directory does not exist, it "
        "will be created",
    )
    parser.add_argument(
        "--channel",
        type=int,
        default=-1,
        help="Channel to draw audio from. Default is to assume mono",
    )
    parser.add_argument(
        "--preprocess",
        type=_json_type,
        default=tuple(),
        help="JSON list of configurations for "
        "'pydrobert.speech.pre.PreProcessor' objects. Audio will be preprocessed in "
        "the same order as the list",
    )
    parser.add_argument(
        "--postprocess",
        type=_json_type,
        default=tuple(),
        help="JSON List of configurations for "
        "'pydrobert.speech.post.PostProcessor' objects. Features will be postprocessed "
        "in the same order as the list",
    )
    parser.add_argument(
        "--force-as",
        default=None,
        choices={
            "table",
            "wav",
            "hdf5",
            "npy",
            "npz",
            "pt",
            "sph",
            "kaldi",
            "file",
            "soundfile",
        }
        | config.SOUNDFILE_SUPPORTED_FILE_TYPES,
        help="Force the paths in 'map' to be interpreted as a specific type "
        "of data. table: kaldi table (key is utterance id); wav: wave file; "
        "hdf5: HDF5 archive (key is utterance id); npy: Numpy binary; npz: "
        "numpy archive (key is utterance id); pt: PyTorch binary; sph: NIST "
        "SPHERE file; kaldi: kaldi object; file: numpy.fromfile binary. soundfile: "
        "force soundfile processing.",
    )
    parser.add_argument(
        "--seed",
        type=_nonneg_int_type,
        default=None,
        help="A random seed used for determinism. This affects operations "
        "like dithering. If unset, a seed will be generated at the moment",
    )
    parser.add_argument(
        "--file-prefix", default="", help="The file prefix indicating a torch data file"
    )
    parser.add_argument(
        "--file-suffix",
        default=".pt",
        help="The file suffix indicating a torch data file",
    )
    parser.add_argument(
        "--num-workers",
        type=_nonneg_int_type,
        default=0,
        help="The number of workers simultaneously computing features. Should not "
        "affect determinism when used in tandem with --seed. '0' means all work is "
        "done on the main thread",
    )
    return parser.parse_args(args)


def signals_to_torch_feat_dir(args=None):
    """Convert a map of signals to a torch SpectDataSet

    This command serves to process audio signals and convert them into a format that can
    be leveraged by "SpectDataSet" in "pydrobert-pytorch"
    (https://github.com/sdrobert/pydrobert-pytorch). It reads in a text file of
    format

        <utt_id_1> <path_to_signal_1>
        <utt_id_2> <path_to_signal_2>
        ...

    computes features according to passed-in settings, and stores them in the
    target directory as

        dir/
            <file_prefix><utt_id_1><file_suffix>
            <file_prefix><utt_id_2><file_suffix>
            ...

    Each signal is read using the utility "pydrobert.speech.util.read_signal()", which
    is a bit slow, but very robust to different file types (such as wave files, hdf5,
    numpy binaries, or Pytorch binaries). A signal is expected to have shape (C, S),
    where C is some number of channels and S is some number of samples. The
    signal can have shape (S,) if the flag "--channels" is set to "-1".

    Features are output as "torch.FloatTensor" of shape "(T, F)", where "T" is some
    number of frames and "F" is some number of filters.

    No checks are performed to ensure that read signals match the feature computer's
    sampling rate (this info may not even exist for some sources).
    """
    try:
        options = _signals_to_torch_feat_dir_parse_args(args)
    except SystemExit as ex:
        return ex.code
    try:
        import torch
        import torch.utils.data
    except ImportError:
        print(
            "signals-to-torch-feat-dir requires a PyTorch installation", file=sys.stderr
        )
        return 1
    if options.seed is None:
        seed = np.random.randint(np.iinfo(np.int32).max)
    else:
        seed = options.seed
    utt2path = dict()
    for line_no, line in enumerate(options.map):
        line = line.strip()
        if not line:
            continue
        ls = line.split(" ")
        if len(ls) < 2:
            print(
                "Line {} of {}: not of format <utt_id> <path>".format(
                    line_no + 1, options.map.name
                ),
                file=sys.stderr,
            )
            return 1
        utt_id = ls[0]
        if utt_id in utt2path:
            print(
                'Line {} of {}: "{}" already exists as utterance'.format(
                    line_no + 1, options.map.name, utt_id
                ),
                file=sys.stderr,
            )
            return 1
        utt2path[utt_id] = " ".join(ls[1:])
    if options.computer_config is None:
        computer = None
    else:
        computer = alias_factory_subclass_from_arg(
            FrameComputer, options.computer_config
        )
    preprocessors = []
    if isinstance(options.preprocess, dict):
        preprocessors.append(
            alias_factory_subclass_from_arg(PreProcessor, options.preprocess)
        )
    else:
        for element in options.preprocess:
            preprocessors.append(alias_factory_subclass_from_arg(PreProcessor, element))
    postprocessors = []
    if isinstance(options.postprocess, dict):
        postprocessors.append(
            alias_factory_subclass_from_arg(PostProcessor, options.postprocess)
        )
    else:
        for element in options.postprocess:
            postprocessors.append(
                alias_factory_subclass_from_arg(PostProcessor, element)
            )
    # we use a dataset to take advantage of torch's multiprocessing
    dataset = _FeatureProcessorDataset(
        utt2path,
        preprocessors,
        computer,
        postprocessors,
        options.channel,
        options.force_as,
        seed,
    )
    loader = torch.utils.data.DataLoader(dataset, num_workers=options.num_workers)
    if not os.path.isdir(options.dir):
        os.makedirs(options.dir)
    for utt_ids, feats in loader:
        utt_id, feat = utt_ids[0], feats[0]
        torch.save(
            feat,
            os.path.join(
                options.dir, options.file_prefix + utt_id + options.file_suffix
            ),
        )
    return 0



import os

from json import load as json_load
from pickle import load as pickle_load

import numpy as np
import pydrobert.speech.filters as filters
import pytest

from pydrobert.speech.config import EFFECTIVE_SUPPORT_THRESHOLD
from pydrobert.speech.alias import alias_factory_subclass_from_arg


@pytest.fixture(
    params=[1, 11], ids=["1 filt", "11 filts"], scope="module",
)
def num_filts(request):
    return request.param


@pytest.fixture(
    params=[
        lambda num_filts: filters.TriangularOverlappingFilterBank(
            "mel", low_hz=5, num_filts=num_filts, sampling_rate=8000, analytic=True,
        ),
        lambda num_filts: filters.TriangularOverlappingFilterBank(
            "mel", low_hz=0, num_filts=num_filts, sampling_rate=8000, analytic=False,
        ),
        lambda num_filts: filters.Fbank(
            low_hz=0, num_filts=num_filts, sampling_rate=8000, analytic=True,
        ),
        lambda num_filts: filters.Fbank(
            low_hz=0, num_filts=num_filts, sampling_rate=8000, analytic=False,
        ),
        lambda num_filts: filters.GaborFilterBank(
            "mel", low_hz=0, num_filts=num_filts, sampling_rate=8000, erb=True,
        ),
        lambda num_filts: filters.GaborFilterBank(
            "mel", low_hz=0, num_filts=num_filts, sampling_rate=8000, erb=False,
        ),
        lambda num_filts: filters.ComplexGammatoneFilterBank(
            "mel",
            low_hz=0,
            num_filts=num_filts,
            sampling_rate=8000,
            max_centered=True,
            erb=True,
        ),
        lambda num_filts: filters.ComplexGammatoneFilterBank(
            "mel",
            low_hz=0,
            num_filts=num_filts,
            sampling_rate=8000,
            max_centered=True,
            erb=False,
        ),
    ],
    ids=[
        "triangular_analytic",
        "triangular",
        "fbank_analytic",
        "fbank",
        "gabor_erb",
        "gabor",
        "gammatone_erb",
        "gammatone",
    ],
)
def bank(request, num_filts):
    return request.param(num_filts)


def test_truncated_matches_full(bank):
    for filt_idx in range(bank.num_filts):
        left_hz, right_hz = bank.supports_hz[filt_idx]
        left_samp, right_samp = bank.supports[filt_idx]
        dft_size = int(
            max(
                (right_samp - left_samp),
                2 * bank.sampling_rate / (right_hz - left_hz),
                1,
            )
        )
        full_response = bank.get_frequency_response(filt_idx, dft_size)
        bin_idx, truncated = bank.get_truncated_response(filt_idx, dft_size)
        challenge = np.zeros(dft_size, dtype=truncated.dtype)
        wrap = min(bin_idx + len(truncated), dft_size) - bin_idx
        challenge[bin_idx : bin_idx + wrap] = truncated[:wrap]
        challenge[: len(truncated) - wrap] = truncated[wrap:]
        if bank.is_real:
            challenge[
                len(challenge)
                - bin_idx
                - len(truncated)
                + 1 : len(challenge)
                - bin_idx
                + 1
            ] = truncated[: None if bin_idx else 0 : -1].conj()
        bad_idx = np.where(
            np.logical_not(
                np.isclose(full_response, challenge, atol=EFFECTIVE_SUPPORT_THRESHOLD)
            )
        )
        assert np.allclose(
            full_response, challenge, atol=EFFECTIVE_SUPPORT_THRESHOLD,
        ), "idx: {} threshold:{} full:{} challenge:{}".format(
            filt_idx,
            EFFECTIVE_SUPPORT_THRESHOLD,
            abs(full_response[bad_idx]),
            abs(challenge[bad_idx]),
        )


def test_frequency_matches_impulse(bank):
    for filt_idx in range(bank.num_filts):
        left_hz, right_hz = bank.supports_hz[filt_idx]
        left_samp, right_samp = bank.supports[filt_idx]
        required_freq_size = 2 * bank.sampling_rate / (right_hz - left_hz)
        required_temp_size = right_samp - left_samp
        if required_temp_size < 5 or required_freq_size < 5:
            # FIXME(sdrobert): this is a stopgap for when filters are
            # *too* localized in time or frequency. This'll cause too
            # much attenuation/gain in one domain or the other.
            continue
        dft_size = int(
            max(
                # allow over- or under-sampling
                (right_samp - left_samp),
                2 * bank.sampling_rate / (right_hz - left_hz),
            )
        )
        X = bank.get_frequency_response(filt_idx, dft_size)
        x = bank.get_impulse_response(filt_idx, dft_size)
        # the absolute tolerance is so high because spectral leakage
        # will muck with the isometry. .001 is enough to say we're in
        # the right ballpark... probably
        assert np.allclose(np.fft.ifft(X), x, atol=1e-3), (len(x), filt_idx)


def test_half_response_matches_full(bank):
    for filt_idx in range(bank.num_filts):
        dft_size = bank.supports[filt_idx][1] - bank.supports[filt_idx][0]
        Xh = bank.get_frequency_response(filt_idx, dft_size, half=True)
        X = bank.get_frequency_response(filt_idx, dft_size, half=False)
        assert np.allclose(X[: len(Xh)], Xh)


def test_zero_outside_freq_support(bank):
    for filt_idx in range(bank.num_filts):
        left_hz, right_hz = bank.supports_hz[filt_idx]
        dft_size = int(max(1, 2 * bank.sampling_rate / (right_hz - left_hz)))
        left_period = int(np.floor(left_hz / bank.sampling_rate))
        right_period = int(np.ceil(right_hz / bank.sampling_rate))
        if right_period - left_period > 2:
            continue
        zero_mask = np.ones(dft_size, dtype=bool)
        for period in range(left_period, right_period + 1):
            for idx in range(dft_size):
                freq = (idx / dft_size + period) * bank.sampling_rate
                zero_mask[idx] &= (freq < left_hz) | (freq > right_hz)
        if bank.is_real:
            zero_mask[1:] &= zero_mask[-1:0:-1]
        if not np.any(zero_mask):
            continue
        X = bank.get_frequency_response(filt_idx, dft_size)
        assert np.allclose(
            X[zero_mask],
            0,
            atol=(right_period - left_period) * EFFECTIVE_SUPPORT_THRESHOLD,
        )


def test_zero_outside_temp_support(bank):
    for filt_idx in range(bank.num_filts):
        left_samp, right_samp = bank.supports[filt_idx]
        width = int(max(1, right_samp - left_samp))
        left_period = int(np.floor(left_samp / width))
        right_period = int(np.ceil(right_samp / width))
        if right_period - left_period > 2:
            continue
        zero_mask = np.ones(width, dtype=bool)
        for period in range(left_period, right_period + 1):
            for idx in range(width):
                t = idx + period * width
                zero_mask[idx] &= (t < left_samp) | (t > right_samp)
        if not np.any(zero_mask):
            continue
        x = bank.get_impulse_response(filt_idx, width)
        assert np.allclose(
            x[zero_mask],
            0,
            atol=(right_period - left_period) * EFFECTIVE_SUPPORT_THRESHOLD,
        )


@pytest.mark.parametrize("window_size", [10, 100, 1000])
@pytest.mark.parametrize("peak_ratio", [0.5, 0.75, 0.9])
@pytest.mark.parametrize("order", [2, 4])
def test_gamma_window_peak_matches(window_size, peak_ratio, order):
    expected_max_idx = window_size * peak_ratio
    window = filters.GammaWindow(order=order, peak=peak_ratio).get_impulse_response(
        window_size
    )
    # if expected_max_idx is a fraction, we don't know what side the peak will
    # fall on a priori because the gamma function does not have symmetric
    # curvature
    max_idx = np.argmax(window)
    assert int(expected_max_idx) == max_idx or int(expected_max_idx) == max_idx + 1


def test_kaldi_filters_match_fbank_filters():
    data_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")
    with open(os.path.join(data_dir, "kaldi_filts.pkl"), "rb") as pkl_file:
        kaldi_40 = pickle_load(pkl_file)
    with open(os.path.join(data_dir, "fbank.json"), "r") as json_file:
        json = json_load(json_file)["bank"]
    bank = alias_factory_subclass_from_arg(filters.LinearFilterBank, json)
    for filt_idx, (kaldi_offset, kaldi_filt) in enumerate(kaldi_40):
        bank_offset, bank_filt = bank.get_truncated_response(filt_idx, 2 ** 9)
        bank_filt **= 2  # see Fbank docs for more info why we do this
        assert bank_offset == kaldi_offset
        assert np.allclose(bank_filt[: len(kaldi_filt)], kaldi_filt, atol=1e-5)
        assert np.allclose(bank_filt[len(kaldi_filt) :], 0.0)
